# -*- coding: utf-8 -*-

"""
Test all support math functions
"""

from __future__ import print_function, division, absolute_import

import math
import cmath
import collections
from itertools import chain

import numba as nb
import numpy as np
from numpy.testing import assert_almost_equal

# ______________________________________________________________________
# Common

def run_common(mod, x):
    "np, math and cmath"
    y0  = mod.sin(x)
    y1  = mod.cos(x)
    y2  = mod.tan(x)
    y3  = mod.sqrt(x)
    y4  = mod.sinh(x)
    y5  = mod.cosh(x)
    y6  = mod.tanh(x)
    y7  = mod.log(x)
    y8  = mod.log10(x)
    y9  = mod.exp(x)
    return (y0, y1, y2, y3, y4, y5, y6, y7, y8, y9)

def run_np_math(mod, x):
    "np (floating, complex) and math (floating)"
    if hasattr(mod, 'expm1'):
        y0  = mod.expm1(x)
    else:
        y0 = 0.0
    y1  = mod.log1p(x)
    return (y0, y1)

def run_commonf(mod, x):
    "np and math"
    y0  = mod.floor(x)
    y1  = mod.ceil(x)
    y2  = mod.hypot(x, x)
    return (y0, y1, y2)

# ______________________________________________________________________
# NumPy

def run_np_arc(mod, x):
    "np only"
    y0  = mod.arccos(x)
    y1  = mod.arcsin(x)
    y2  = mod.arctan(x)
    y3  = mod.arcsinh(x)
    y4  = mod.arccosh(1.0/x)
    return (y0, y1, y2, y3, y4)

def run_np_misc(mod, x):
    "np only"
    y0  = mod.log2(x)
    y1  = mod.exp2(x)
    y2  = mod.rint(x)
    y3  = mod.power(x, x)
    y4  = mod.absolute(x * -1) # TODO: USub for some types
    return (y0, y1, y2, y3, y4)

# ______________________________________________________________________
# Python

def run_py_arc(mod, x):
    "math and cmath"
    y0  = mod.acos(x)
    y1  = mod.asin(x)
    y2  = mod.atan(x)
    y3  = mod.asinh(x)
    y4  = mod.acosh(1.0/x)
    y5  = mod.atanh(x)
    return (y0, y1, y2, y3, y4, y5)

def misc_floating(mod, x):
    "miscellaneous"
    # y0  = math.erfc(x)

    y1  = math.atan2(x, x)
    y2  = np.arctan2(x, x)
    y3  = np.logaddexp(x, x)
    y4  = np.logaddexp2(x, x)
    return (y1, y2, y3, y4) #(y0, y1)

# ______________________________________________________________________
# Run tests

Suite = collections.namedtuple('Suite', ['mod', 'types'])
merge = lambda d1, d2: dict(chain(d1.items(), d2.items()))

integral = nb.short, nb.int_, nb.uint, nb.long_, nb.ulong, nb.longlong, nb.ulonglong
floating = nb.float_, nb.double, #nb.longdouble
complexes = nb.complex64, nb.complex128, #nb.complex256

fdata     = { integral : 6, floating: 6.0, (nb.object_,): 6.0 }
cdata     = { complexes: 6.0+4.0j }
data      = merge(fdata, cdata)

arc_fdata = { floating: 0.6, (nb.object_,): 0.6 }
arc_cdata = { complexes: 0.6+0.4j }
arc_data  = merge(arc_fdata, arc_cdata)

tests = {
    run_common  : [Suite(math, fdata), Suite(cmath, cdata)],
    run_np_math : [Suite(np, data), Suite(math, fdata)],
    run_commonf : [Suite(np, fdata), Suite(math, fdata)],
    run_np_arc  : [Suite(np, arc_data)],
    run_py_arc  : [Suite(math, arc_fdata), Suite(cmath, arc_cdata)],
    run_np_misc : [Suite(np, data)],
    # run_py_math : [Suite(math, fdata)],
    misc_floating : [Suite(math, fdata)]
}

def run():
    for test, suites in tests.iteritems():
        for suite in suites:
            for types, data in suite.types.iteritems():
                for ty in types:
                    print("running:", test.__name__, ty)
                    signature = nb.object_(nb.typeof(suite.mod), ty)
                    jitter = nb.jit(signature)
                    jitted = jitter(test)

                    r1 = test(suite.mod, data)
                    r2 = jitted(suite.mod, data)
                    # print(r1, r2)
                    assert np.allclose(r1, r2)

                    r3 = jitted(suite.mod, data)
                    assert np.allclose(r1, r3)
run()



import os
from .generator import build

root = os.path.dirname(os.path.abspath(__file__))
features = ['ast', 'transformer', 'visitor']

def build_normalized():
    fn = os.path.join(root, "Normalized.asdl")
    outdir = os.path.join(root, "normalized")
    build.build_package(fn, features, outdir)

def build_untyped():
    fn = os.path.join(root, "UntypedIR.asdl")
    outdir = os.path.join(root, "untyped")
    build.build_package(fn, features, outdir)

def build_typed():
    fn = os.path.join(root, "TypedIR.asdl")
    outdir = os.path.join(root, "typed")
    build.build_package(fn, features, outdir)


if __name__ == '__main__':
    build_normalized()
    build_untyped()
    # build_typed()


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import os
import sys
import shutil
import tempfile
import subprocess

from .. import naming, build

root = os.path.dirname(os.path.abspath(__file__))
schema_filename = os.path.join(root, "testschema1.asdl")

def filenames(codegens):
    return [c.out_filename for c in codegens]

all_features = [naming.interface + '.pxd', naming.nodes + '.pxd',
                naming.visitor + '.pxd', naming.transformer + '.pxd',
                naming.interface + '.py',  naming.nodes + '.py',
                naming.visitor + '.py', naming.transformer + '.py',]

def test_features():
    """
    >>> test_features()
    """
    codegens = build.enumerate_codegens(['ast'], mask=0)
    fns = filenames(codegens)
    assert set(fns) == set([naming.interface + '.pxd', naming.nodes + '.pxd',
                            naming.interface + '.py',  naming.nodes + '.py',])

    codegens = build.enumerate_codegens(['ast'], mask=build.cython)
    fns = filenames(codegens)
    assert set(fns) == set([naming.interface + '.py',  naming.nodes + '.py',])

    codegens = build.enumerate_codegens(
        ['ast', 'visitor', 'transformer'], mask=0)
    fns = filenames(codegens)
    assert set(fns) == set(all_features)

def test_package_building():
    """
    >>> test_package_building()
    """
    features = ['ast', 'visitor', 'transformer']
    output_dir = tempfile.mkdtemp()
    try:
        build.build_package(schema_filename, features, output_dir)
        for feature_file in all_features:
            assert os.path.exists(os.path.join(output_dir, feature_file))
    finally:
        shutil.rmtree(output_dir)

def test_package_compilation():
    """
    >>> test_package_compilation()
    """
    features = ['ast', 'visitor', 'transformer']
    output_dir = tempfile.mkdtemp()
    try:
        build.build_package(schema_filename, features, output_dir)
        p = subprocess.Popen([sys.executable, "setup.py",
                              "build_ext", "--inplace"], cwd=output_dir)
        assert p.wait() == 0, p.poll()
    finally:
        shutil.rmtree(output_dir)


if __name__ == '__main__':
    import doctest
    try:
        import cython
    except ImportError:
        print("Skipping test, cython not installed")
    else:
        sys.exit(0 if doctest.testmod().failed == 0 else 1)


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
# For reference:
#    typedef struct {
#    PyObject_HEAD                   // indices (skipping the head)
#    char *data;                     // 0
#    int nd;                         // 1
#    int *dimensions, *strides;      // 2, 3
#    PyObject *base;                 // 4
#    PyArray_Descr *descr;           // 5
#    int flags;                      // 6
#    } PyArrayObject;

import abc

from numba import *
from numba import typedefs
from numba.typesystem import tbaa
from numba.llvm_types import _head_len, _int32, _LLVMCaster, constant_int
import llvm.core as lc

def _const_int(X):
    return lc.Constant.int(lc.Type.int(), X)

def ptr_at(builder, ptr, idx):
    return builder.gep(ptr, [_const_int(idx)])

def load_at(builder, ptr, idx):
    return builder.load(ptr_at(builder, ptr, idx))

def store_at(builder, ptr, idx, val):
    builder.store(val, ptr_at(builder, ptr, idx))


def set_metadata(tbaa, instr, type):
    if type is not None:
        metadata = tbaa.get_metadata(type)
        instr.set_metadata("tbaa", metadata)

def make_property(type=None, invariant=True):
    """
    type: The type to be used for TBAA annotation
    """
    def decorator(access_func):
        def load(self):
            instr = self.builder.load(access_func(self))
            if self.tbaa:
                set_metadata(self.tbaa, instr, type)
            return instr

        def store(self, value):
            ptr = access_func(self)
            instr = self.builder.store(value, ptr)
            if self.tbaa:
                set_metadata(self.tbaa, instr, type)

        return property(load, store)

    return decorator

class PyArrayAccessor(object):
    """
    Convenient access to a the native fields of a NumPy array.

    builder: llvmpy IRBuilder
    pyarray_ptr: pointer to the numpy array
    tbaa: metadata.TBAAMetadata instance
    """

    def __init__(self, builder, pyarray_ptr, tbaa=None, dtype=None):
        self.builder = builder
        self.pyarray_ptr = pyarray_ptr
        self.tbaa = tbaa # this may be None
        self.dtype = dtype

    def _get_element(self, idx):
        indices = [constant_int(0), constant_int(_head_len + idx)]
        ptr = self.builder.gep(self.pyarray_ptr, indices)
        return ptr

    def get_data(self):
        instr = self.builder.load(self._get_element(0))
        if self.tbaa:
            set_metadata(self.tbaa, instr, self.dtype.pointer())
        return instr

    def set_data(self, value):
        instr = self.builder.store(value, self._get_element(0))
        if self.tbaa:
            set_metadata(self.tbaa, instr, self.dtype.pointer())

    data = property(get_data, set_data, "The array.data attribute")

    def typed_data(self, context):
        data = self.data
        ltype = self.dtype.pointer().to_llvm(context)
        return self.builder.bitcast(data, ltype)

    @make_property(tbaa.numpy_ndim)
    def ndim(self):
        return self._get_element(1)

    @make_property(tbaa.numpy_shape.pointer().qualify("const"))
    def dimensions(self):
        return self._get_element(2)

    shape = dimensions

    @make_property(tbaa.numpy_strides.pointer().qualify("const"))
    def strides(self):
        return self._get_element(3)

    @make_property(tbaa.numpy_base)
    def base(self):
        return self._get_element(4)

    @make_property(tbaa.numpy_dtype)
    def descr(self):
        return self._get_element(5)

    @make_property(tbaa.numpy_flags)
    def flags(self):
        return self._get_element(6)


class Array(object):
    """
    Interface for foreign arrays, like LLArray
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def from_type(cls, llvm_dtype):
        """
        Given an LLVM representation of the dtype, return the LLVM array type
        representation
        """

    @abc.abstractproperty
    def data(self):
        """Return the data pointer of this array (for A.data)"""

    @abc.abstractproperty
    def shape_ptr(self):
        """Return the shape pointer of this array (for A.shape[0], etc)"""

    @abc.abstractproperty
    def strides_ptr(self):
        """Return the strides pointer of this array (for A.strides[0], etc)"""

    @abc.abstractproperty
    def shape(self):
        """Return the extents as a list of loaded LLVM values"""

    @abc.abstractproperty
    def strides(self):
        """Return the strides as a list of loaded LLVM values"""

    @abc.abstractproperty
    def ndim(self):
        """Return the dimensionality of this array as an LLVM constant"""

    @abc.abstractmethod
    def getptr(self, *indices):
        """Compute an element pointer given LLVM indices into the array"""


class NumpyArray(Array):
    """
    LLArray compatible inferface for NumPy's ndarray
    """

    _strides_ptr = None
    _strides = None
    _shape_ptr = None
    _shape = None
    _data_ptr = None
    _freefuncs = []
    _freedata = []

    def __init__(self, pyarray_ptr, builder, tbaa=None, type=None):
        self.type = type
        self.nd = type.ndim
        self.array_type = pyarray_ptr.type.pointee

        # LLVM attributes
        self.arr = PyArrayAccessor(builder, pyarray_ptr, tbaa, type.dtype)
        self.builder = builder
        self._shape = None
        self._strides = None
        self.caster = _LLVMCaster(builder)

    @classmethod
    def from_type(cls, llvm_dtype):
        return typedefs.PyArray.pointer().to_llvm()

    @property
    def data(self):
        if not self._data_ptr:
            self._data_ptr = self.arr.get_data()
        return self._data_ptr

    @property
    def shape_ptr(self):
        if self._shape_ptr is None:
            self._shape_ptr = self.arr.shape
        return self._shape_ptr

    @property
    def strides_ptr(self):
        if self._strides_ptr is None:
            self._strides_ptr = self.arr.strides
        return self._strides_ptr

    @property
    def shape(self):
        if not self._shape:
            self._shape = self.preload(self.shape_ptr, self.nd)
        return self._shape

    @property
    def strides(self):
        if not self._strides:
            self._strides = self.preload(self.strides_ptr, self.nd)
        return self._strides

    @property
    def ndim(self):
        return _const_int(self.nd)

    def getptr(self, *indices):
        offset = _const_int(0)
        for i, (stride, index) in enumerate(zip(self.strides, indices)):
            index = self.caster.cast(index, stride.type, unsigned=False)
            offset = self.caster.cast(offset, stride.type, unsigned=False)
            offset = self.builder.add(offset, self.builder.mul(index, stride))

        data_ty = self.type.dtype.to_llvm()
        data_ptr_ty = lc.Type.pointer(data_ty)

        dptr_plus_offset = self.builder.gep(self.data, [offset])

        ptr = self.builder.bitcast(dptr_plus_offset, data_ptr_ty)
        return ptr

    # Misc, optional methods

    @property
    def itemsize(self):
        raise NotImplementedError

    def preload(self, ptr, count=None):
        assert count is not None
        return [load_at(self.builder, ptr, i) for i in range(count)]


# -*- coding: UTF-8 -*-
from __future__ import print_function, division, absolute_import

from io import StringIO

from numba import config
from numba.annotate import render_text
from numba.annotate.annotate import (Source, Annotation, SourceIntermediate,
                                     Program, A_type)

# ______________________________________________________________________

py_source = Source(
    linemap={ 1: u'def foo(a, b):',
              2: u'    print a * b',
              3: u'    a / b',
              4: u'    return a - b' },
    annotations={ 2: [Annotation(A_type, (u'a', u'double')),
                      Annotation(A_type, (u'b', u'double'))] }
)

linenomap = { 1: [0], 2: [1, 2, 3], 4: [5], }
llvm_linemap = {
    0: u'call @printf(%a, %b)',
    1: u'%0 = load a',
    2: u'%1 = load b',
    3: u'%2 = fadd %0 %1',
    4: u'%3 = fdiv %a %b',
    5: u'ret something',
}
annotations = {
    3: [Annotation(A_type, (u'%0', u'double')),
        Annotation(A_type, (u'%1', u'double'))],
}
llvm_intermediate = SourceIntermediate("llvm", linenomap,
                                       Source(llvm_linemap, annotations))
# p = Program(py_source, [llvm_intermediate])
p = [{'python_source': py_source, 'intermediates': [llvm_intermediate]}]

# ______________________________________________________________________

config.config.colour = False    # Disable lexing for tests

def run_render_text():
    f = StringIO()
    render_text(p, emit=f.write)
    src = f.getvalue()
    assert 'def foo(a, b):' in src, src
    assert 'print a * b' in src
    assert 'return a - b' in src
    assert 'double' in src
    # print(src)

def run_render_text_inline():
    f = StringIO()
    render_text(p, emit=f.write, intermediate_names=["llvm"])
    src = f.getvalue()
    assert 'def foo(a, b):' in src
    assert '____llvm____' in src
    assert '%0 = load a' in src
    # print(src)

def run_render_text_outline():
    f = StringIO()
    render_text(p, emit=f.write, inline=False, intermediate_names=["llvm"])
    src = f.getvalue()
    assert 'def foo(a, b):' in src
    assert "====llvm====" in src
    assert '%0 = load a' in src
    # print(src)

run_render_text()
run_render_text_inline()
run_render_text_outline()



from numba import autojit
from numba.testing.test_support import autojit_py3doc

@autojit_py3doc
def max1(x):
    """
    >>> max1([100])
    100
    >>> max1([1,2.0,3])
    3
    >>> max1([-1,-2,-3.0])
    -1
    >>> max1(1)
    Traceback (most recent call last):
        ...
    TypeError: 'int' object is not iterable
    """
    return max(x)

@autojit_py3doc
def min1(x):
    """
    >>> min1([100])
    100
    >>> min1([1,2,3.0])
    1
    >>> min1([-1,-2.0,-3])
    -3
    >>> min1(1)
    Traceback (most recent call last):
        ...
    TypeError: 'int' object is not iterable
    """
    return min(x)

@autojit_py3doc
def max2(x, y):
    """
    >>> max2(1, 2)
    2
    >>> max2(1, -2)
    1
    >>> max2(10, 10.25)
    10.25
    >>> max2(10, 9.9)
    10.0
    >>> max2(0.1, 0.25)
    0.25
    >>> max2(1, 'a')
    Traceback (most recent call last):
        ...
    UnpromotableTypeError: Cannot promote types int and string
    """
    return max(x, y)

@autojit_py3doc
def min2(x, y):
    """
    >>> min2(1, 2)
    1
    >>> min2(1, -2)
    -2
    >>> min2(10, 10.1)
    10.0
    >>> min2(10, 9.75)
    9.75
    >>> min2(0.25, 0.3)
    0.25
    >>> min2(1, 'a')
    Traceback (most recent call last):
        ...
    UnpromotableTypeError: Cannot promote types int and string
    """
    return min(x, y)

@autojit
def max4(x):
    """
    >>> max4(20)
    20.0
    """
    return max(1, 2.0, x, 14)

@autojit
def min4(x):
    """
    >>> min4(-2)
    -2.0
    """
    return min(1, 2.0, x, 14)


if __name__ == '__main__':
    import numba
    numba.testing.testmod()



from . import _common

from numba import llvm_types

from llvm_cbuilder import *
from llvm_cbuilder import shortnames as C

from numba.codegen.llvmcontext import LLVMContextManager
from numba.vectorize import _internal
from numba import decorators

import numpy as np


class _GeneralizedUFuncFromFunc(_common.CommonVectorizeFromFunc):
    def datalist(self, lfunclist, ptrlist):
        """
        Return a list of data pointers to the kernels.
        """
        return [None] * len(lfunclist)

    def __call__(self, lfunclist, tyslist, signature, engine,
                 vectorizer, **kws):
        '''create generailized ufunc from a llvm.core.Function

        lfunclist : a single or iterable of llvm.core.Function instance
        engine : a llvm.ee.ExecutionEngine instance

        return a function object which can be called from python.
        '''
        assert 'cuda_dispatcher' not in kws, "Temporary check for mismatch API"
        kws['signature'] = signature

        try:
            iter(lfunclist)
        except TypeError:
            lfunclist = [lfunclist]

        self.tyslist = tyslist
        ptrlist = self._prepare_pointers(lfunclist, tyslist, engine, **kws)
        inct = len(tyslist[0]) - 1
        outct = 1
        datlist = self.datalist(lfunclist, ptrlist)

        # Becareful that fromfunc does not provide full error checking yet.
        # If typenum is out-of-bound, we have nasty memory corruptions.
        # For instance, -1 for typenum will cause segfault.
        # If elements of type-list (2nd arg) is tuple instead,
        # there will also memory corruption. (Seems like code rewrite.)

        # Hold on to the vectorizer while the ufunc lives
        tyslist = self.get_dtype_nums(tyslist)
        gufunc = _internal.fromfuncsig(ptrlist, tyslist, inct, outct, datlist,
                                       signature, vectorizer)

        return gufunc

    def build(self, lfunc, dtypes, signature):
        def_guf = GUFuncEntry(dtypes, signature, CFuncRef(lfunc))
        guf = def_guf(lfunc.module)
        # print guf
        return guf


class GUFuncASTVectorize(object):
    """
    Vectorizer for generalized ufuncs.
    """

    def __init__(self, func, sig):
        self.pyfunc = func
        self.translates = []
        self.signature = sig
        self.gufunc_from_func = _GeneralizedUFuncFromFunc()
        self.args_restypes = getattr(self, 'args_restypes', [])
        self.signatures = []
        self.llvm_context = LLVMContextManager()

    def _get_lfunc_list(self):
        return [t.lfunc for t in self.translates]

    def build_ufunc(self):
        assert self.translates, "No translation"
        lfunclist = self._get_lfunc_list()
        tyslist = self._get_tys_list()
        engine = self._get_ee()
        return self.gufunc_from_func(
            lfunclist, tyslist, self.signature, engine,
            vectorizer=self)

    def get_argtypes(self, numba_func):
        return numba_func.signature.args

    def _get_ee(self):
        return self.llvm_context.execution_engine

    def add(self, restype=None, argtypes=None):
        dec = decorators.jit(restype, argtypes, backend='ast')
        numba_func = dec(self.pyfunc)
        self.args_restypes.append(list(numba_func.signature.args) +
                                  [numba_func.signature.return_type])
        self.signatures.append((restype, argtypes, {}))
        self.translates.append(numba_func)

    def _get_tys_list(self):
        types_lists = []
        for numba_func in self.translates:
            dtype_nums = []
            types_lists.append(dtype_nums)
            for arg_type in self.get_argtypes(numba_func):
                if arg_type.is_array:
                    arg_type = arg_type.dtype
                dtype_nums.append(arg_type.get_dtype())

        return types_lists

GUFuncVectorize = GUFuncASTVectorize

_intp_ptr = C.pointer(C.intp)

class PyObjectHead(CStruct):
    _fields_ = [
        ('ob_refcnt', C.intp),
        # NOTE: not a integer, just need to match definition in numba
        ('ob_type', C.pointer(C.int)),
    ]

    if llvm_types._trace_refs_:
        # Account for _PyObject_HEAD_EXTRA
        _fields_ = [
            ('ob_next', _intp_ptr),
            ('ob_prev', _intp_ptr),
        ] + _fields_


class PyArray(CStruct):

    _fields_ = PyObjectHead._fields_ + [
        ('data',           C.void_p),
        ('nd',             C.int),
        ('dimensions',     _intp_ptr),
        ('strides',        _intp_ptr),
        ('base',           C.void_p),
        ('descr',          C.void_p),
        ('flags',          C.int),
        ('weakreflist',    C.void_p),
#        ('maskna_dtype',   C.void_p),
#        ('maskna_data',    C.void_p),
#        ('maskna_strides', _intp_ptr),
    ]

    def fakeit(self, dtype, data, dimensions, steps):
        assert len(dimensions) == len(steps)
        constant = self.parent.constant

        self.ob_refcnt.assign(constant(C.intp, 1))
        type_p = constant(C.py_ssize_t, id(np.ndarray))
        self.ob_type.assign(type_p.cast(C.pointer(C.int)))

        self.base.assign(self.parent.constant_null(C.void_p))
        dtype_p = constant(C.py_ssize_t, id(dtype))
        self.descr.assign(dtype_p.cast(C.void_p))
        self.flags.assign(constant(C.int, _internal.NPY_WRITEABLE))

        self.data.assign(data)
        self.nd.assign(constant(C.int, len(dimensions)))

        ary_dims = self.parent.array(C.intp, len(dimensions) * 2)
        ary_steps = ary_dims[len(dimensions):]
        for i, dim in enumerate(dimensions):
            ary_dims[i] = dim

        self.dimensions.assign(ary_dims)

        # ary_steps = self.parent.array(C.intp, len(steps))
        for i, step in enumerate(steps):
            ary_steps[i] = step
        self.strides.assign(ary_steps)


def _parse_signature(sig):
    inargs, outarg = sig.split('->')

    for inarg in filter(bool, inargs.split(')')):
        dimnames = inarg[1+inarg.find('('):].split(',')
        yield dimnames
    else:
        dimnames = outarg.strip('()').split(',')
        yield dimnames

class GUFuncEntry(CDefinition):
    '''a generalized ufunc that wraps a numba jit'ed function

    NOTE: Currently, this only works for array return type.
    And, return type must be the last argument of the nubma jit'ed function.
    '''
    _argtys_ = [
        ('args',       C.pointer(C.char_p)),
        ('dimensions', C.pointer(C.intp)),
        ('steps',      C.pointer(C.intp)),
        ('data',       C.void_p),
    ]

    def _outer_loop(self, args, dimensions, pyarys, steps, data):
        # implement outer loop
        innerfunc = self.depends(self.FuncDef)
        with self.for_range(dimensions[0]) as (loop, idx):
            args = [arg.reference().cast(arg_type.type)
                        for arg, arg_type in zip(pyarys, innerfunc.handle.args)]
            innerfunc(*args)
            # innerfunc(*map(lambda x: x.reference(), pyarys))

            for i, ary in enumerate(pyarys):
                ary.data.assign(ary.data[steps[i]:])

    def body(self, args, dimensions, steps, data):
        diminfo = list(_parse_signature(self.Signature))
        n_pyarys = len(diminfo)
        assert n_pyarys == len(self.dtypes)

        # extract unique dimension names
        dims = []
        for grp in diminfo:
            for it in grp:
                if it not in dims:
                    dims.append(it)

        # build pyarrays for argument to inner function
        pyarys = [self.var(PyArray) for _ in range(n_pyarys)]

        # populate pyarrays
        step_offset = len(pyarys)
        for i, (dtype, ary) in enumerate(zip(self.dtypes, pyarys)):
            ary_ndim = len(diminfo[i])
            ary_dims = [dimensions[1 + dims.index(k)] for k in diminfo[i]]
            ary_steps = []

            for j in range(ary_ndim):
                ary_steps.append(steps[step_offset])
                step_offset += 1

            ary.fakeit(dtype, args[i], ary_dims, ary_steps)

        self._outer_loop(args, dimensions, pyarys, steps, data)
        self.ret()

    @classmethod
    def specialize(cls, dtypes, signature, func_def):
        '''specialize to a workload
        '''
        signature = signature.replace(' ', '') # remove all spaces
        cls.dtypes = dtypes
        cls._name_ = 'gufunc_%s_%s'% (signature, func_def)
        cls.FuncDef = func_def
        cls.Signature = signature



# -*- coding: utf-8 -*-

"""
ASDL processor, handles imports. This should be part of the parser,
but we have many multiple copies of those...
"""

from __future__ import print_function, division, absolute_import

import os
import re
import tokenize

from numba.asdl.asdl import ASDLProcessor

#------------------------------------------------------------------------
# Process ASDL imports
#------------------------------------------------------------------------

class ImportProcessor(ASDLProcessor):

    def __init__(self, asdlmod, import_path):
        self.asdlmod = asdlmod
        self.import_path = import_path

    def preprocess(self, asdl_source):
        # Find and save import statements. Remove imports from source
        self.imports, source = find_imports(asdl_source)
        return source

    def postprocess(self, asdl_tree):
        # Import subtrees and types
        subtrees, types = apply_imports(self.asdlmod,
                                        self.imports,
                                        self.import_path)

        # Merge imported subtrees and types
        asdl_tree.dfns.extend(subtrees)
        asdl_tree.types.update(types)

        return asdl_tree

#------------------------------------------------------------------------
# Find Source Imports
#------------------------------------------------------------------------

pattern = "^\s*from (\w+) import (\*|\w+(?:, \w+)*)$"

def find_imports(asdl_source):
    """
    Find imports in the given ASDL source.

    :return: Two-tuple of (imports, source) where source has the imports
             removed and imports is a list of two-tuples (modname, (names))

             ([(str, (str,))], str)
    """
    lines = asdl_source.splitlines()
    imports = []

    source_lines = []
    for line in lines:
        m = re.match(pattern, line)
        if m is not None:
            module_name, names = m.groups()
            if names == '*':
                import_tuple = (module_name, (names,))
            else:
                import_tuple = (module_name, tuple(names.split(", ")))

            imports.append(import_tuple)
        else:
            source_lines.append(line)

    return imports, "\n".join(source_lines)

#------------------------------------------------------------------------
# Find Imported Terms
#------------------------------------------------------------------------

def apply_imports(asdlmod, imports, import_path):
    """
    Import ASDL subtrees from another schema along the given import path.
    """
    from . import asdl

    subtrees = []
    types = {}

    for modname, import_names in imports:
        # Find module file
        for path in import_path:
            fname = os.path.join(path, modname + '.asdl')
            if os.path.exists(fname):
                # Load ASDL tree
                parser, loader = asdl.load(modname, open(fname).read(), asdlmod)
                tree = loader.load()

                # Collect subtrees and types by name
                handle_import(tree, import_names, subtrees, types)
                break
        else:
            raise ImportError("No module named %r" % (modname + '.asdl',))

    return subtrees, types

def handle_import(tree, import_names, subtrees, types):
    for import_name in import_names:
        dfns = from_import(tree, import_name)
        subtrees.extend(dfns)
        for dfn in dfns:
            print(dfn.name, tree.types.keys())
            types[dfn.name] = tree.types[str(dfn.name)]

def from_import(tree, import_name):
    if import_name == '*':
        return tree.dfns #[dfn for dfn in tree.dfns if dfn.name in tree.types]

    for definition in tree.dfns:
        if str(definition.name) == import_name:
            return [definition]

    raise ImportError("Module %r has no rule %r" % (tree.name, import_name))



# -*- coding: utf-8 -*-

"""
Visualize an SSA graph (the def/use chains).
"""

from __future__ import print_function, division, absolute_import

import os
import textwrap

from numba.viz.graphviz import render
from numba.viz.cfgviz import cf_from_source
from numba.control_flow.cfstats import NameAssignment

# ______________________________________________________________________
# Adaptor

class SSAGraphAdaptor(object):
    def children(self, node):
        return [nameref.variable for nameref in node.cf_references]

# ______________________________________________________________________
# Renderer

class SSAGraphRenderer(object):

    def render(self, node):
        if node.renamed_name:
            return node.unmangled_name
        return node.name

    def render_edge(self, source, dest):
        return "use"

# ______________________________________________________________________
# Entry Points

def render_ssa(cfflow, symtab, output_file):
    "Render the SSA graph given the flow.CFGFlow and the symbol table"
    cfstats = [stat for b in cfflow.blocks for stat in b.stats]
    defs = [stat.lhs.variable for stat in cfstats
                                  if isinstance(stat, NameAssignment)]
    nodes = symtab.values() + defs
    render(nodes, output_file, SSAGraphAdaptor(), SSAGraphRenderer())

def render_ssa_from_source(source, output_file, func_globals=()):
    "Render the SSA graph given python source code"
    symtab, cfflow = cf_from_source(source, func_globals)
    render_ssa(cfflow, symtab, output_file)

# ______________________________________________________________________
# Test

if __name__ == '__main__':
    source = textwrap.dedent("""
        def func():
            # x_0
            x = 0 # x_1
            # x_2
            for i in range(10):
                if i < 5:
                    x = i # x_3
                # x_4
                # x = x + i # x_5

            y = x
            x = i # x_6
    """)

    render_ssa_from_source(source, os.path.expanduser("~/ssa.dot"))


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba.typesystem import numba_typesystem as ts

def test_pointers():
    p = ts.pointer(ts.pointer(ts.int_))
    assert str(p) == "int **", str(p)

def test_arrays():
    A = ts.array(ts.double, 1)
    B = ts.array(ts.double, 2)

    assert str(A) == "float64[:]"
    assert str(A[1:]) == "float64"
    assert str(B[1:]) == "float64[:]"
    assert str(B[-1:10]) == "float64[:]"
    assert str(B[0:]) == "float64[:, :]"
    assert str(B[0:10]) == "float64[:, :]"
    assert str(B[-2:10]) == "float64[:, :]"

def test_functions():
    functype = ts.function(ts.int_, (ts.float_,))
    assert str(functype) == "int (*)(float32)", functype
    functype = ts.function(ts.int_, (ts.float_,), "hello")
    assert str(functype) == "int (*hello)(float32)", functype

if __name__ == "__main__":
    test_pointers()
    test_arrays()
    test_functions()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

class Config(object):
    # Add color to printed source code
    colour = True

    # Terminal background colour ("light" or "dark")
    terminal_background = "dark"

config = Config()


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import ast
import textwrap
try:
    import __builtin__ as builtins
except ImportError:
    import builtins

import numba
from numba import missing
from numba import *
from numba import error
from numba import typesystem
from numba import visitors, nodes
from numba.typesystem import get_type
from numba.specialize import loopimpl

logger = logging.getLogger(__name__)

#------------------------------------------------------------------------
# Utilities
#------------------------------------------------------------------------

def unpack_range_args(node):
    start, stop, step = (nodes.const(0, Py_ssize_t),
                         None,
                         nodes.const(1, Py_ssize_t))

    if len(node.args) == 0:
        raise error.NumbaError(node, "Expected at least one argument")
    elif len(node.args) == 1:
        stop, = node.args
    elif len(node.args) == 2:
        start, stop = node.args
    else:
        start, stop, step = node.args

    return [start, stop, step]

def make_while_loop(flow_node):
    "Create a while loop from a flow node (a While or If node)"
    while_node = nodes.While(test=flow_node.test,
                             body=flow_node.body,
                             orelse=flow_node.orelse)
    return ast.copy_location(while_node, flow_node)

def copy_basic_blocks(flow_node_src, flow_node_dst):
    "Copy cfg basic blocks from one flow node to another"
    flow_node_dst.cond_block = flow_node_src.cond_block
    flow_node_dst.if_block   = flow_node_src.if_block
    flow_node_dst.else_block = flow_node_src.else_block
    flow_node_dst.exit_block = flow_node_src.exit_block

def make_while_from_for(for_node):
    "Create a While from a For. The 'test' (loop condition) must still be set."
    while_node = nodes.While(test=None,
                             body=for_node.body,
                             orelse=for_node.orelse)
    copy_basic_blocks(for_node, while_node)
    while_node = nodes.build_while(**vars(while_node))
    return ast.copy_location(while_node, for_node)

def untypedTemp():
    "Temp node with a yet unknown type"
    type = typesystem.DeferredType(None)
    temp = nodes.TempNode(type)
    type.variable = temp.variable
    return temp

#------------------------------------------------------------------------
# Transform for loops
#------------------------------------------------------------------------

class TransformForIterable(visitors.NumbaTransformer):
    """
    This transforms loops over 1D arrays and loops over range().
    """

    def rewrite_range_iteration(self, node):
        """
        Handle range iteration:

            for i in range(start, stop, step):
                ...

        becomes

            nsteps = compute_nsteps(start, stop, step)
            temp = 0

            while temp < nsteps:
                target = start + temp * step
                ...
                temp += 1
        """
        self.generic_visit(node)

        temp = nodes.TempNode(node.target.type, 'target_temp')
        nsteps = nodes.TempNode(Py_ssize_t, 'nsteps')
        start, stop, step = unpack_range_args(node.iter)

        if isinstance(step, nodes.ConstNode):
            have_step = step.pyval != 1
        else:
            have_step = True

        start, stop, step = map(nodes.CloneableNode, (start, stop, step))

        if have_step:
            templ = textwrap.dedent("""
                    {{temp}} = 0
                    {{nsteps}} = ({{stop}} - {{start}} + {{step}} -
                                    (1 if {{step}} >= 0 else -1)) / {{step}}
                    while {{temp_load}} < {{nsteps_load}}:
                        {{target}} = {{start}} + {{temp_load}} * {{step}}
                        {{temp}} = {{temp_load}} + 1
                        {{body}}
                """)
        else:
            templ = textwrap.dedent("""
                    {{temp}} = {{start}}
                    {{nsteps}} = {{stop}}
                    while {{temp_load}} < {{nsteps_load}}:
                        {{target}} = {{temp_load}}
                        {{temp}} = {{temp_load}} + 1
                        {{body}}
                """)

        if node.orelse:
            templ += "\nelse: {{else_body}}"

        # Leave the bodies empty, they are already analyzed
        body = ast.Suite(body=[])
        else_body = ast.Suite(body=[])

        #--------------------------------------------------------------------
        # Substitute template and infer types
        #--------------------------------------------------------------------

        result = self.run_template(
            templ, vars=dict(length=Py_ssize_t),
            start=start, stop=stop, step=step,
            nsteps=nsteps.store(), nsteps_load=nsteps.load(),
            temp=temp.store(), temp_load=temp.load(),
            target=node.target,
            body=body, else_body=else_body)
        ast.copy_location(result, node)
        if hasattr(node, 'lineno'):
            visitor = missing.FixMissingLocations(node.lineno, node.col_offset,
                                              override=True)
            visitor.visit(result)

        #--------------------------------------------------------------------
        # Patch the body and else clause
        #--------------------------------------------------------------------

        body.body.extend(node.body)
        else_body.body.extend(node.orelse)

        while_node = result.body[-1]
        assert isinstance(while_node, ast.While)

        #--------------------------------------------------------------------
        # Create a While with the ForNode's cfg blocks merged in
        #--------------------------------------------------------------------

        while_node = make_while_loop(while_node)
        copy_basic_blocks(node, while_node)
        while_node = nodes.build_while(**vars(while_node))

        # Create the place to jump to for 'continue'
        while_node.continue_block = node.cond_block

        # Set the new while loop in the templated Suite
        result.body[-1] = while_node

        return result

    def rewrite_array_iteration(self, node):
        """
        Convert 1D array iteration to for-range and indexing:

            for value in my_array:
                ...

        becomes

            for i in my_array.shape[0]:
                value = my_array[i]
                ...
        """
        logger.debug(ast.dump(node))

        orig_target = node.target
        orig_iter = node.iter

        #--------------------------------------------------------------------
        # Replace node.target with a temporary
        #--------------------------------------------------------------------

        target_temp = nodes.TempNode(typesystem.Py_ssize_t)
        node.target = target_temp.store()

        #--------------------------------------------------------------------
        # Create range(A.shape[0])
        #--------------------------------------------------------------------

        call_func = ast.Name(id='range', ctx=ast.Load())
        nodes.typednode(call_func, typesystem.range_)

        shape_index = ast.Index(nodes.ConstNode(0, typesystem.Py_ssize_t))
        shape_index.type = typesystem.npy_intp

        stop = ast.Subscript(value=nodes.ShapeAttributeNode(orig_iter),
                             slice=shape_index,
                             ctx=ast.Load())
        nodes.typednode(stop, npy_intp)

        #--------------------------------------------------------------------
        # Create range iterator and replace node.iter
        #--------------------------------------------------------------------

        call_args = [nodes.ConstNode(0, typesystem.Py_ssize_t),
                     nodes.CoercionNode(stop, typesystem.Py_ssize_t),
                     nodes.ConstNode(1, typesystem.Py_ssize_t),]

        node.iter = ast.Call(func=call_func, args=call_args)
        nodes.typednode(node.iter, call_func.type)

        node.index = target_temp.load(invariant=True)

        #--------------------------------------------------------------------
        # Add assignment to new target variable at the start of the body
        #--------------------------------------------------------------------

        index = ast.Index(value=node.index)
        index.type = target_temp.type
        subscript = ast.Subscript(value=orig_iter,
                                  slice=index, ctx=ast.Load())
        nodes.typednode(subscript, get_type(orig_iter).dtype)

        #--------------------------------------------------------------------
        # Add assignment to new target variable at the start of the body
        #--------------------------------------------------------------------

        assign = ast.Assign(targets=[orig_target], value=subscript)
        node.body = [assign] + node.body

        #--------------------------------------------------------------------
        # Specialize new for loop through range iteration
        #--------------------------------------------------------------------

        return self.visit(node)

    def visit_For(self, node):
        if node.iter.type.is_range:
            return self.rewrite_range_iteration(node)
        elif node.iter.type.is_array and node.iter.type.ndim == 1:
            return self.rewrite_array_iteration(node)
        else:
            self.visitchildren(node)
            return node

#------------------------------------------------------------------------
# Transform for loops over builtins
#------------------------------------------------------------------------

class TransformBuiltinLoops(visitors.NumbaTransformer):
    def rewrite_enumerate(self, node):
        """
        Rewrite a loop like

        for i, x in enumerate(array[, start]):
            ...

        into

        _arr = array
        [_s = start]
        for _i in range(len(_arr)):
            i = _i [+ _s]
            x = _arr[_i]
            ...
        """
        call = node.iter
        if (len(call.args) not in (1, 2) or call.keywords or
            call.starargs or call.kwargs):
            self.error(call, 'expected 1 or 2 arguments to enumerate()')

        target = node.target
        if (not isinstance(target, (ast.Tuple, ast.List)) or
            len(target.elts) != 2):
            self.error(call, 'expected 2 iteration variables')

        array = call.args[0]
        start = call.args[1] if len(call.args) > 1 else None
        idx = target.elts[0]
        var = target.elts[1]

        array_temp = untypedTemp()
        if start:
            start_temp = untypedTemp() # TODO: only allow integer start
        idx_temp = nodes.TempNode(typesystem.Py_ssize_t)

        # for _i in range(len(_arr)):
        node.target = idx_temp.store()
        node.iter = ast.Call(ast.Name('range', ast.Load()),
                             [ast.Call(ast.Name('len', ast.Load()),
                                       [array_temp.load(True)],
                                       [], None, None)],
                             [], None, None)

        # i = _i [+ _s]
        new_idx = idx_temp.load()
        if start:
            new_idx = ast.BinOp(new_idx, ast.Add(), start_temp.load(True))
        node.body.insert(0, ast.Assign([idx], new_idx))

        # x = _arr[_i]
        value = ast.Subscript(array_temp.load(True),
                              ast.Index(idx_temp.load()),
                              ast.Load())
        node.body.insert(1, ast.Assign([var], value))

        # _arr = array; [_s = start]; ...
        body = [ ast.Assign([array_temp.store()], array), node ]
        if start:
            body.insert(1, ast.Assign([start_temp.store()], start))
        return map(self.visit, body)

    def rewrite_zip(self, node):
        """
        Rewrite a loop like

        for x, y... in zip(xs, ys...):
            ...

        into

        _xs = xs; _ys = ys...
        for _i in range(min(len(_xs), len(_ys)...)):
            x = _xs[_i]; y = _ys[_i]...
            ...
        """
        call = node.iter
        if not call.args or call.keywords or call.starargs or call.kwargs:
            self.error(call, 'expected at least 1 argument to zip()')

        target = node.target
        if (not isinstance(target, (ast.Tuple, ast.List)) or
            len(target.elts) != len(call.args)):
            self.error(call, 'expected %d iteration variables' % len(call.args))

        temps = [untypedTemp() for _ in xrange(len(call.args))]
        idx_temp = nodes.TempNode(typesystem.Py_ssize_t)

        # min(len(_xs), len(_ys)...)
        len_call = ast.Call(ast.Name('min', ast.Load()),
                            [ast.Call(ast.Name('len', ast.Load()),
                                      [tmp.load(True)], [], None, None)
                             for tmp in temps],
                            [], None, None)

        # for _i in range(...):
        node.target = idx_temp.store()
        node.iter = ast.Call(ast.Name('range', ast.Load()),
                             [len_call], [], None, None)

        # x = _xs[_i]; y = _ys[_i]...
        node.body = [ast.Assign([tgt],
                                ast.Subscript(tmp.load(True),
                                              ast.Index(idx_temp.load()),
                                              ast.Load()))
                     for tgt, tmp in zip(target.elts, temps)] + \
                    node.body

        # _xs = xs; _ys = ys...
        body = [ast.Assign([tmp.store()], arg)
                for tmp, arg in zip(temps, call.args)] + \
               [node]
        return map(self.visit, body)

    HANDLERS = {
        id(enumerate): rewrite_enumerate,
        id(zip):       rewrite_zip,
    }

    def visit_For(self, node):
        if (isinstance(node.iter, ast.Call) and
            isinstance(node.iter.func, ast.Name)):
            name = node.iter.func.id
            if name not in self.symtab:
                obj = (self.func_globals[name]
                       if name in self.func_globals else
                       getattr(builtins, name, None))
                rewriter = self.HANDLERS.get(id(obj))
                if rewriter:
                    return rewriter(self, node)

        self.visitchildren(node)
        return node

#------------------------------------------------------------------------
# Transform for loops over Objects
#------------------------------------------------------------------------

class SpecializeObjectIteration(visitors.NumbaTransformer):
    """
    This transforms for loops over objects.
    """

    def visit_For(self, node):
        while_node = make_while_from_for(node)

        test = nodes.const(True, bool_)
        while_node.test = test

        impl = loopimpl.find_iterator_impl(node)

        # Get the iterator, loop body, and the item
        iter = impl.getiter(self.context, node, self.llvm_module)
        body = impl.body(self.context, node, self.llvm_module)
        item = impl.next(self.context, node, self.llvm_module)

        # Coerce item to LHS and assign
        item = nodes.CoercionNode(item, node.target.type)
        target_assmnt = ast.Assign(targets=[node.target], value=item)

        # Update While node body
        body.insert(0, target_assmnt)
        while_node.body = body

        nodes.merge_cfg_in_while(while_node)

        return ast.Suite(body=[iter, while_node])



# -*- coding: utf-8 -*-

"""
Universes of type constructors for numba.
"""

from __future__ import print_function, division, absolute_import

import struct as struct_module
import ctypes

import numpy as np

names = lambda *names: list(names) #list(map(tyname, names))

int_typenames = names(
    'char', 'uchar', 'short', 'ushort', 'int', 'uint', 'long', 'ulong',
    'longlong', 'ulonglong', 'int8', 'int16', 'int32', 'int64',
    'uint8', 'uint16', 'uint32', 'uint64',
    'size_t', 'npy_intp', 'Py_ssize_t', 'Py_uintptr_t',
    'bool', # hmm
)

signed = frozenset(names(
    'char', 'short', 'int', 'long', 'longlong',
    'int8', 'int16', 'int32', 'int64',
    'Py_ssize_t', 'npy_intp',
))

float_typenames = names(
    'float', 'double', 'float32', 'float64', #'longdouble', 'float128',
)

complex_typenames = names(
    'complex64', 'complex128', #'complex256',
)

#------------------------------------------------------------------------
# Default type sizes
#------------------------------------------------------------------------

_plat_bits = struct_module.calcsize('@P') * 8

def getsize(ctypes_name, default):
    try:
        return ctypes.sizeof(getattr(ctypes, ctypes_name))
    except ImportError:
        return default

# Type sizes in bytes
type_sizes = {
    "bool":         1,
    # Int
    "char":         1,
    "int8":         1,
    "int16":        2,
    "int32":        4,
    "int64":        8,
    # Unsigned int
    "uchar":        1,
    "uint8":        1,
    "uint16":       2,
    "uint32":       4,
    "uint64":       8,
    # Float
    # "float16":      2,
    "float32":      4,
    "float64":      8,
    # "float128":     16,
    "float":        4,
    "double":       8,
    # Complex
    "complex64":    8,
    "complex128":   16,
    # "complex256":   32,
}

ctypes_npy_intp = np.empty(0).ctypes.strides._type_

sizeof_longdouble = np.dtype(np.longdouble).itemsize # Use numpy's opinion here

native_sizes = {
    "char":         1,
    "uchar":        1,
    # Int
    "short":        struct_module.calcsize("h"),
    "int":          struct_module.calcsize("i"),
    "long":         struct_module.calcsize("l"),
    "longlong":     struct_module.calcsize("Q"),
    "Py_ssize_t":   getsize('c_size_t', _plat_bits // 8),
    "npy_intp":     ctypes.sizeof(ctypes_npy_intp),
    # Unsigned int
    "ushort":       struct_module.calcsize("H"),
    "uint":         struct_module.calcsize("I"),
    "ulong":        struct_module.calcsize("L"),
    "ulonglong":    struct_module.calcsize("Q"),
    "size_t":       getsize('c_size_t', _plat_bits // 8),
    "Py_uintptr_t": ctypes.sizeof(ctypes.c_void_p),
    # Float
    # ctypes and numpy may disagree on longdouble
    # "longdouble":   sizeof_longdouble,
    # "float128":     sizeof_longdouble,
    # Complex
    # "complex256":   sizeof_longdouble * 2,

    # Pointer
    "pointer":      ctypes.sizeof(ctypes.c_void_p),
}

default_type_sizes = dict(type_sizes, **native_sizes)

is_native_int = native_sizes.__contains__


# -*- coding: utf-8 -*-

"""
Pretty printing of numba IRs.
"""

from __future__ import print_function, division, absolute_import

import os
import sys

from numba.lexing import lex_source
from numba.viz import cfgviz, astviz
from numba.annotate import annotators
from numba.annotate import render_text, render_html
from numba.annotate.annotate import Source, Program, build_linemap

# ______________________________________________________________________

def dumppass(option):
    def decorator(f):
        def wrapper(ast, env):
            if env.cmdopts.get(option):
                f(ast, env, env.cmdopts.get("fancy"))
            return ast
        return wrapper
    return decorator

# ______________________________________________________________________

@dumppass("dump-ast")
def dump_ast(ast, env, fancy):
    if fancy:
        astviz.render_ast(ast, os.path.expanduser("~/ast.dot"))
    else:
        import ast as ast_module
        print(ast_module.dump(ast))

@dumppass("dump-cfg")
def dump_cfg(ast, env, fancy):
    cfg = env.crnt.flow
    if fancy:
        cfgviz.render_cfg(cfg, os.path.expanduser("~/cfg.dot"))
    else:
        for block in cfg.blocks:
            print(block)
            print("    ", block.parents)
            print("    ", block.children)

@dumppass("annotate")
def dump_annotations(ast, env, fancy):
    llvm_intermediate, = [i for i in env.crnt.intermediates if i.name == "llvm"]
    annotators.annotate_pyapi(llvm_intermediate, env.crnt.annotations)

    p = Program(Source(build_linemap(env.crnt.func), env.crnt.annotations),
                env.crnt.intermediates)
    if fancy:
        render = render_html
        fn, ext = os.path.splitext(env.cmdopts["filename"])
        out = open(fn + '.html', 'w')
        print("Writing", fn + '.html')
    else:
        render = render_text
        out = sys.stdout

    # Look back through stack until we find the line of code that called our
    # jitted function.
    func_call = ''
    func_call_filename = env.cmdopts['filename']
    func_call_lineno = ''
    import traceback
    stack = traceback.extract_stack()
    for i in range(len(stack)-1, -1, -1):
        if stack[i][0] == env.cmdopts['filename'] and stack[i][3].find(env.crnt.func_name) > -1:
            func_call = stack[i][3]
            func_call_lineno = str(stack[i][1])
            break

    annotation = {'func_call':func_call,
                  'func_call_filename':func_call_filename,
                  'func_call_lineno':func_call_lineno,
                  'python_source':p.python_source,
                  'intermediates':p.intermediates}
    if fancy:
        env.annotation_blocks.append(annotation)
    else:
        env.annotation_blocks = [annotation]

    render(env.annotation_blocks, emit=out.write, intermediate_names=["llvm"])

@dumppass("dump-llvm")
def dump_llvm(ast, env, fancy):
    print(lex_source(str(env.crnt.lfunc), "llvm", "console"))

@dumppass("dump-optimized")
def dump_optimized(ast, env, fancy):
    print(lex_source(str(env.crnt.lfunc), "llvm", "console"))


# -*- coding: utf-8 -*-

"""
Visualize an AST.
"""

from __future__ import print_function, division, absolute_import

import os
import ast
import textwrap
from itertools import chain, imap, ifilter

from numba.viz.graphviz import render

# ______________________________________________________________________
# AST Constants

ast_constant_classes = (
    ast.expr_context,
    ast.operator,
    ast.unaryop,
    ast.cmpop,
)

# ______________________________________________________________________
# Utilities

is_ast = lambda node: (isinstance(node, (ast.AST, list)) and not
                       isinstance(node, ast_constant_classes))

class SomeConstant(object):
    def __init__(self, value):
        self.value = value
    def __repr__(self):
        return repr(self.value)

def make_list(node):
    if isinstance(node, list):
        return node
    elif isinstance(node, ast.AST):
        return [node]
    else:
        return [SomeConstant(node)]

def nodes(node):
    return [getattr(node, attr, None) for attr in node._fields]

def fields(node):
    return zip(node._fields, nodes(node))

# ______________________________________________________________________
# Adaptor

class ASTGraphAdaptor(object):

    def children(self, node):
        return list(chain(*imap(make_list, ifilter(is_ast, nodes(node)))))

# ______________________________________________________________________
# Renderer

def strval(val):
    if isinstance(val, ast_constant_classes):
        return type(val).__name__ # Load, Store, Param
    else:
        return repr(val)

class ASTGraphRenderer(object):

    def render(self, node):
        all_fields = fields(node)

        for attr in getattr(node, '_attributes', []):
            if attr not in node._fields and hasattr(node, attr):
                all_fields.append((attr, getattr(node, attr)))

        all_fields = [(attr, v) for attr, v in all_fields if not is_ast(v)]
        args = ",\n".join('%s=%s' % (a, strval(v)) for a, v in all_fields)
        if args:
            args = '\n' + args
        return "%s(%s)" % (type(node).__name__, args)

    def render_edge(self, source, dest):
        # See which attribute of the source node matches the destination node
        for attr_name, attr in fields(source):
            if attr is dest or (isinstance(attr, list) and dest in attr):
                # node.attr == dst_node or dest_node in node.attr
                return attr_name

# ______________________________________________________________________
# Entry Point

def render_ast(ast, output_file):
    render([ast], output_file, ASTGraphAdaptor(), ASTGraphRenderer())

# ______________________________________________________________________
# Test

if __name__ == '__main__':
    source = textwrap.dedent("""
        def func(a, b):
            for i in range(10):
                if i < 5:
                    print "hello"
    """)
    mod = ast.parse(source)
    print(mod)
    render_ast(mod, os.path.expanduser("~/ast.dot"))


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

# Thanks to Neal Becker

# TODO: Fix and dedup

import numpy as np
from numba import *

f8_array_ty = f8[:]

@autojit
class fir (object):
    ## @void(f8[:])
    def __init__ (self, coef):
        self.coef = coef
        self.inp = np.zeros_like (coef)

    ## @void(f8)
    def shift (self, u):
        size = self.inp.size
        for i in range (size-1-1, -1, -1):
            self.inp[i+1] = self.inp[i]
        self.inp[0] = u

    ## @double ()
    def compute1(self):
        s = 0
        size = self.coef.size
        for i in range (size):
            s += self.inp[i] * self.coef[i]
        return s

    ## @f8(f8)
    def shift_compute1 (self, u):
        self.shift (u)
        return self.compute1()

    ## @f8_array_ty(f8_array_ty)
    def compute (self, u):
        out = np.empty_like (u)
        size = u.size
        for i in range (size):
            out[i] = self.shift_compute1 (u[i])
        return out

    ## @f8_array_ty(f8_array_ty)
    def __call__ (self, u):
        return self.compute (u)

if __name__ == '__main__':
    from timeit import timeit

    coef = np.arange(100, dtype=np.double)
    filt = fir (coef)

    inp = np.arange(100, dtype=np.double)
    out = filt.compute (inp)

    filt (coef)




# -*- coding: utf-8 -*-

"""
User-facing numba types.
"""

from __future__ import print_function, division, absolute_import

import numba

import ctypes
from itertools import imap
from functools import partial

from numba import odict
from numba.typesystem.itypesystem import Type, Conser, add_flags, tyname

import numpy as np

#------------------------------------------------------------------------
# Type metaclass
#------------------------------------------------------------------------

numba_type_registry = odict.OrderedDict()
register = numba_type_registry.__setitem__

class Accessor(object):
    def __init__(self, idx):
        self.idx = idx

    def __get__(self, obj, type=None):
        return obj.params[self.idx]

    def __set__(self, obj, value):
        if not obj.mutable:
            raise AttributeError("Cannot set attribute '%s' of type '%s'" %
                                            (obj.argnames[self.idx], type(obj)))
        obj.params[self.idx] = value

class TypeMetaClass(type):
    "Metaclass for numba types, conses immutable types."

    def __init__(self, name, bases, dict):
        if dict.get('typename') is None and name[0].islower():
            self.typename = name.rstrip("_")
        if self.typename is not None:
            register(self.typename, self)
        _update_class(self)
        self.conser = Conser(partial(type.__call__, self))

    def __call__(self, *args, **kwds):
        args = self.default_args(args, kwds)
        if not self.mutable:
            return self.conser.get(*args)
        return type.__call__(self, *args)

#------------------------------------------------------------------------
# Type Decorators
#------------------------------------------------------------------------

def _update_class(cls):
    # Build defaults dict { argname : default_value }
    if 'defaults' not in vars(cls):
        cls.defaults = {}
        for i, argname in enumerate(cls.argnames):
            if isinstance(argname, (list, tuple)):
                name, default = argname
                cls.argnames[i] = name
                cls.defaults[name] = default

    # Create accessors
    for i, arg in enumerate(vars(cls).get("argnames", ())):
        assert not getattr(cls, arg, False), (cls, arg)
        setattr(cls, arg, Accessor(i))

    # Process flags
    flags = list(cls.flags)
    if cls.typename:
        flags.append(cls.typename.strip("_"))

    add_flags(cls, flags)

def consing(cls):
    """
    Cons calls to the constructor.
    """
    cls.mutable = False
    return cls

def notconsing(cls):
    cls.mutable = True
    return cls

#------------------------------------------------------------------------
# Type Implementations
#------------------------------------------------------------------------

class _NumbaType(Type):
    """
    MonoType with user-facing methods:

        call: create a function type
        slice: create an array type
        conversion: to_llvm/to_ctypes/get_dtype
    """

    argnames = []
    flags = []
    defaults = {}

    qualifiers = frozenset()

    # ______________________________________________________________________
    # Internal

    def add(self, attr, value):
        "Construct new type with attr=value (e.g. functype.add('args', []))"
        assert not self.mutable
        params = list(self.params)
        params[self.argnames.index(attr)] = value
        return type(self)(*params)

    @property
    def subtypes(self):
        subtypes = []
        for p in self.params:
            if isinstance(p, (Type, list, tuple)):
                subtypes.append(p)

        return subtypes

    # ______________________________________________________________________
    # User functionality

    def pointer(self):
        return pointer(self)

    def ref(self):
        return reference(self)

    def qualify(self, *qualifiers):
        return self # TODO: implement

    def unqualify(self, *qualifiers):
        return self # TODO: implement

    # TODO: Remove context argument in favour of typesystem argument
    def to_llvm(self, context=None):
        from . import defaults
        return defaults.numba_typesystem.convert("llvm", self)

    def to_ctypes(self):
        from . import defaults
        return defaults.numba_typesystem.convert("ctypes", self)

    def to_numpy(self):
        from numba.typesystem import numpy_support
        return numpy_support.to_dtype(self)

    get_dtype = to_numpy

    # ______________________________________________________________________
    # Special methods (user functionality)

    def __getitem__(self, item):
        """
        Support array type creation by slicing, e.g. double[:, :] specifies
        a 2D strided array of doubles. The syntax is the same as for
        Cython memoryviews.
        """
        assert isinstance(item, (tuple, slice)), item

        def verify_slice(s):
            if s.start or s.stop or s.step not in (None, 1):
                raise ValueError(
                    "Only a step of 1 may be provided to indicate C or "
                    "Fortran contiguity")

        if isinstance(item, tuple):
            step_idx = None
            for idx, s in enumerate(item):
                verify_slice(s)
                if s.step and (step_idx or idx not in (0, len(item) - 1)):
                    raise ValueError(
                        "Step may only be provided once, and only in the "
                        "first or last dimension.")

                if s.step == 1:
                    step_idx = idx

            return array_(self, len(item),
                             is_c_contig=step_idx == len(item) - 1,
                             is_f_contig=step_idx == 0)
        else:
            verify_slice(item)
            return array_(self, 1, is_c_contig=bool(item.step))

    def __call__(self, *args):
        """
        Return a new function type when called with type arguments.
        """
        if len(args) == 1 and not isinstance(args[0], Type):
            # Cast in Python space
            # TODO: Create proxy object
            # TODO: Fully customizable type system (do this in Numba, not
            #       minivect)
            return args[0]

        return function(self, args)

@notconsing
class NumbaType(_NumbaType):
    """
    Base for numba types.
    """

    __metaclass__ = TypeMetaClass
    # __slots__ = Type.slots

    typename = None

    def __init__(self, *args, **kwds):
        super(NumbaType, self).__init__(self.typename, *args, **kwds)
        assert len(args) == len(self.argnames), (self.typename, args)
        for name in kwds:
            assert name in self.argnames, (self.typename, kwds, self.argnames)

    @classmethod
    def default_args(cls, args, kwargs):
        names = cls.argnames
        if len(args) == len(names):
            return args

        # Insert defaults in args tuple
        args = list(args)
        for name in names[len(args):]:
            if name in kwargs:
                args.append(kwargs[name])
            elif name in cls.defaults:
                args.append(cls.defaults[name])
            else:
                raise TypeError("Constructor '%s' requires %d arguments "
                                "(got %d)" % (cls.typename, len(names), len(args)))

        return tuple(args)

#------------------------------------------------------------------------
# Low-level parametrized types
#------------------------------------------------------------------------

def pass_by_ref(type): # TODO: Get rid of this
    return type.is_struct or type.is_complex

class Function(object):
    """
    Function types may be called with Python functions to create a Function
    object. This may be used to minivect users for their own purposes. e.g.

    @double(double, double)
    def myfunc(...):
       ...
    """

    def __init__(self, signature, py_func):
        self.signature = signature
        self.py_func = py_func

    def __call__(self, *args, **kwargs):
        """
        Implement this to pass the callable test for classmethod/staticmethod.
        E.g.

            @classmethod
            @void()
            def m(self):
                ...
        """
        raise TypeError("Not a callable function")

@consing
class function(NumbaType):
    typename = "function"
    argnames = ['return_type', 'args', ('name', None), ('is_vararg', False)]

    def add_arg(self, i, arg):
        args = list(self.args)
        args.insert(i, arg)
        return self.add('args', args)

    # ______________________________________________________________________

    @property
    def struct_by_reference(self):
        rt = self.return_type
        byref = lambda t: t.is_struct or t.is_complex
        return rt and byref(rt) or any(imap(byref, self.args))

    @property
    def actual_signature(self):
        """
        Passing structs by value is not properly supported for different
        calling conventions in LLVM, so we take an extra argument
        pointing to a caller-allocated struct value.
        """
        from numba import typesystem as ts

        if self.struct_by_reference:
            args = []
            for arg in self.args:
                if pass_by_ref(arg):
                    arg = arg.pointer()
                args.append(arg)

            return_type = self.return_type
            if pass_by_ref(self.return_type):
                return_type = ts.void
                args.append(self.return_type.pointer())

            self = function(return_type, args)

        return self

    @property
    def struct_return_type(self):
        # Function returns a struct.
        return self.return_type.pointer()

    # ______________________________________________________________________

    def __repr__(self):
        args = [str(arg) for arg in self.args]
        if self.is_vararg:
            args.append("...")
        if self.name:
            namestr = self.name
        else:
            namestr = ''

        return "%s (*%s)(%s)" % (self.return_type, namestr, ", ".join(args))

    def __call__(self, *args):
        if len(args) != 1 or isinstance(args[0], Type):
            return super(function, self).__call__(*args)

        assert self.return_type is not None
        assert self.argnames is not None
        func, = args
        return Function(self, func)

# ______________________________________________________________________
# Pointers

@consing
class pointer(NumbaType):
    argnames = ['base_type']

    @property
    def is_string(self): # HACK
        import numba
        return self.base_type == numba.char

    def __repr__(self):
        space = " " * (not self.base_type.is_pointer)
        return "%s%s*" % (self.base_type, space)

@consing
class sized_pointer(NumbaType):
    """
    A pointer with knowledge of its range.

    E.g. an array's 'shape' or 'strides' attribute.
    This also allow tuple unpacking.
    """
    typename = "sized_pointer"
    argnames = ["base_type", "size"]
    flags = ["pointer"]

    # def __eq__(self, other):
    #     if other.is_sized_pointer:
    #         return (self.base_type == other.base_type and
    #                 self.size == other.size)
    #     return other.is_pointer and self.base_type == other.base_type
    #
    # def __hash__(self):
    #     return hash(self.base_type.pointer())

@consing
class carray(NumbaType):
    argnames = ["base_type", "size"]

# ______________________________________________________________________
# Structs

@consing
class istruct(NumbaType):

    argnames = ["fields", ("name", None), ("readonly", False), ("packed", False)]

    @property
    def subtypes(self):
        return [f[1] for f in self.fields]

    @property
    def fielddict(self):
        return dict(self.fields)

    def __repr__(self):
        if self.name:
            name = self.name + ' '
        else:
            name = ''
        return 'struct %s{ %s }' % (
                name, ", ".join(["%s %s" % (field_type, field_name)
                                    for field_name, field_type in self.fields]))

    def is_prefix(self, other_struct):
        other_fields = other_struct.fields[:len(self.fields)]
        return self.fields == other_fields

    def offsetof(self, field_name):
        """
        Compute the offset of a field. Must be used only after mutation has
        finished.
        """
        ctype = self.to_ctypes()
        return getattr(ctype, field_name).offset

@notconsing
class struct_(istruct):
    """
    Create a struct type. Fields may be ordered or unordered. Unordered fields
    will be ordered from big types to small types (for better alignment).
    """
    mutable = True

    def __eq__(self, other):
        return other.is_struct and self.fields == other.fields

    def __hash__(self):
        return hash(tuple(self.fields))

    def copy(self):
        return type(self)(self.fields, self.name, self.readonly, self.packed)

    def add_field(self, name, type):
        assert name not in self.fielddict
        self.fields.append((name, type))
        self.mutated = True

    def update_mutated(self):
        self.rank = sum([sort_key(field) for field in self.fields])
        self.mutated = False

#------------------------------------------------------------------------
# High-level types
#------------------------------------------------------------------------

@consing
class array_(NumbaType):
    """
    An array type. array_ may be sliced to obtain a subtype:

    >>> double[:, :, ::1][1:]
    double[:, ::1]
    >>> double[:, :, ::1][:-1]
    double[:, :]

    >>> double[::1, :, :][:-1]
    double[::1, :]
    >>> double[::1, :, :][1:]
    double[:, :]
    """

    argnames = ["dtype", "ndim", "is_c_contig", "is_f_contig", "inner_contig"]
    defaults = dict.fromkeys(argnames[-3:], False)
    flags = ["object"]

    def pointer(self):
        raise Exception("You probably want a pointer type to the dtype")

    @property
    def strided(self):
        return array_(self.dtype, self.ndim)

    def __repr__(self):
        axes = [":"] * self.ndim
        if self.is_c_contig and self.ndim > 0:
            axes[-1] = "::1"
        elif self.is_f_contig and self.ndim > 0:
            axes[0] = "::1"

        return "%s[%s]" % (self.dtype, ", ".join(axes))

    def __getitem__(self, index):
        "Slicing an array slices the dimensions"
        assert isinstance(index, slice)
        assert index.step is None
        assert index.start is not None or index.stop is not None

        start = 0
        stop = self.ndim
        if index.start is not None:
            start = index.start
        if index.stop is not None:
            stop = index.stop

        ndim = len(range(self.ndim)[start:stop])

        if ndim == 0:
            return self.dtype
        elif ndim > 0:
            return type(self)(self.dtype, ndim)
        else:
            raise IndexError(index, ndim)

@consing
class autojit_function(NumbaType):
    "Type for autojit functions"
    argnames = ["autojit_func"]
    flags = ["object"]

@consing
class jit_function(NumbaType):
    "Type for jit functions"
    argnames = ["jit_func"]
    flags = ["object"]

@consing
class numpy_dtype(NumbaType):
    "Type of numpy dtypes"
    argnames = ["dtype"]
    flags = ["object"]

@consing
class complex_(NumbaType):
    argnames = ["base_type"]
    flags = ["numeric"]

    @property
    def itemsize(self):
        return self.base_type.itemsize * 2

    def __repr__(self):
        return "complex%d" % (self.itemsize * 8,)

@consing
class meta(NumbaType):
    """
    A type instance in user code. e.g. double(value). The Name node will have
    a cast-type with dst_type 'double'.
    """
    argnames = ["dst_type"]
    flags = [
        "object",
        "cast",     # backwards compat
    ]

#------------------------------------------------------------------------
# Container Types
#------------------------------------------------------------------------

@consing
class ContainerListType(NumbaType):
    """
    :param base_type: the element type of the tuple
    :param size: set to a value >= 0 is the size is known
    :return: a tuple type representation
    """

    argnames = ["base_type", "size"]
    flags = ["object", "container"]

    def is_sized(self):
        return self.size >= 0

@consing
class tuple_(ContainerListType):
    "tuple(base_type, size)"

@consing
class list_(ContainerListType):
    "list(base_type, size)"

@consing
class MapContainerType(NumbaType):
    argnames = ["key_type", "value_type", "size"]
    flags = ["object"]

@consing
class dict_(MapContainerType):
    "dict(key, value, size)"

#------------------------------------------------------------------------
# Types to be removed
#------------------------------------------------------------------------

class numpy_attribute(NumbaType): # TODO: remove
    argnames = ["module", "attr"]
    flags = ["object", "known_value"]

    @property
    def value(self):
        return getattr(self.module, self.attr)

class module_attribute(NumbaType): # TODO: remove
    argnames = ["module", "attr"]
    flags = ["object", "known_value"]

    @property
    def value(self):
        return getattr(self.module, self.attr)

@consing
class reference(NumbaType): # TODO: remove ?
    """
    A reference to an (primitive or Python) object. This is passed as a
    pointer and dereferences automatically.

    Currently only supported for structs.
    """
    argnames = ["referenced_type"]

@consing
class method(NumbaType): # TODO: remove
    """
    Method of something.

        base_type: the object type the attribute was accessed on
    """
    argnames = ["base_type", "attr_name"]
    flags = ["object"]

class pointer_to_function(NumbaType): # TODO: remove
    """
    Pointer to a function at a known address represented by some Python
        object (e.g. a ctypes or CFFI function).
    """
    typename = "pointer_to_function"
    argnames = ["obj", "ptr", "signature"]
    flags = ["object"]

@consing
class known_value(NumbaType): # TODO: remove
    """
    Type which is associated with a known value or well-defined symbolic
    expression:

        np.add          => np.add
        np.add.reduce   => (np.add, "reduce")

    (Remember that unbound methods like np.add.reduce are transient, i.e.
     np.add.reduce is not np.add.reduce).
    """
    argnames = ["value"]

@consing
class known_pointer(pointer): # TODO: remove
    argnames = ["base_type", "address"]

@notconsing
class global_(known_value): # TODO: Remove
    "Global type"

@consing
class builtin_(known_value): # TODO: remove
    argnames = ["name", "value"]
    flags = ["object"]

    @property
    def func(self):
        return self.value

@consing
class module(known_value): # TODO: remove
    """
    Represents a type for modules.

    Attributes:
        is_numpy_module: whether the module is the numpy module
        module: in case of numpy, the numpy module or a submodule
    """
    flags = ["object"]

    # TODO: Get rid of these
    is_numpy_module = property(lambda self: self.module is np)
    is_numba_module = property(lambda self: self.module is np)
    is_math_module = property(lambda self: self.module is np)

    @property
    def module(self):
        return self.value

#------------------------------------------------------------------------
# Convenience functions...
#------------------------------------------------------------------------

unit = _NumbaType.unit
_array = array_
_struct = struct_

def from_numpy_dtype(np_dtype):
    """
    :param np_dtype: the NumPy dtype (e.g. np.dtype(np.double))
    :return: a dtype type representation
    """
    from numba.typesystem import numpy_support
    return numpy_dtype(numpy_support.map_dtype(np_dtype))

def array(dtype, ndim, is_c_contig=False, is_f_contig=False, inner_contig=False):
    """
    :param dtype: the Numba dtype type (e.g. double)
    :param ndim: the array dimensionality (int)
    :return: an array type representation
    """
    if ndim == 0:
        return dtype
    return _array(dtype, ndim, is_c_contig, is_f_contig, inner_contig)

# ______________________________________________________________________

def sort_key(t):
    n, ty = t
    return ctypes.sizeof(ty.to_ctypes())

def struct_(fields=(), name=None, readonly=False, packed=False, **kwargs):
    "Create a mutable struct type"
    if fields and kwargs:
        raise TypeError("The struct must be either ordered or unordered")
    elif kwargs:
        import ctypes
        fields = sorted(kwargs.iteritems(), key=sort_key, reverse=True)
        # fields = sort_types(kwargs)
        # fields = list(kwargs.iteritems())

    return _struct(fields, name, readonly, packed)


# -*- coding: utf-8 -*-

from __future__ import print_function, division, absolute_import

import math

from numba.typesystem import *
from numba.typesystem.itypesystem import nbo

import numpy as np

def map_dtype(dtype):
    """
    Map a NumPy dtype to a minitype.

    >>> map_dtype(np.dtype(np.int32))
    int32
    >>> map_dtype(np.dtype(np.int64))
    int64
    >>> map_dtype(np.dtype(np.object))
    PyObject *
    >>> map_dtype(np.dtype(np.float64))
    float64
    >>> map_dtype(np.dtype(np.complex128))
    complex128
    """
    if dtype.byteorder not in ('=', nbo, '|') and dtype.kind in ('iufbc'):
        raise minierror.UnmappableTypeError(
                "Only native byteorder is supported", dtype)

    item_idx = int(math.log(dtype.itemsize, 2))
    if dtype.kind == 'i':
        return [int8, int16, int32, int64][item_idx]
    elif dtype.kind == 'u':
        return [uint8, uint16, uint32, uint64][item_idx]
    elif dtype.kind == 'f':
        if dtype.itemsize == 2:
            pass # half floats not supported yet
        elif dtype.itemsize == 4:
            return float32
        elif dtype.itemsize == 8:
            return float64
        elif dtype.itemsize == 16:
            raise TypeError("long double is not support")
            # return float128
    elif dtype.kind == 'b':
        return int8
    elif dtype.kind == 'c':
        if dtype.itemsize == 8:
            return complex64
        elif dtype.itemsize == 16:
            return complex128
        elif dtype.itemsize == 32:
            raise TypeError("long double is not support")
            # return complex256
    elif dtype.kind == 'V':
        fields = [(name, map_dtype(dtype.fields[name][0]))
                      for name in dtype.names]
        is_aligned = dtype.alignment != 1
        return struct_(fields, packed=not getattr(dtype, 'isalignedstruct',
                                                  is_aligned))
    elif dtype.kind == 'O':
        return object_

typemap = {
    int8     : np.int8,
    int16    : np.int16,
    int32    : np.int32,
    int64    : np.int64,
    uint8    : np.uint8,
    uint16   : np.uint16,
    uint32   : np.uint32,
    uint64   : np.uint64,

    float_   : np.float32,
    double   : np.float64,
    # longdouble: np.longdouble,

    short    : np.dtype('h'),
    int_     : np.dtype('i'),
    long_    : np.dtype('l'),
    longlong : np.longlong,
    ushort   : np.dtype('H'),
    uint     : np.dtype('I'),
    ulong    : np.dtype('L'),
    ulonglong: np.ulonglong,

    complex64: np.complex64,
    complex128: np.complex128,
    # complex256: getattr(np, 'complex256', None),

    bool_    : np.bool,
    object_  : np.object,
}
typemap = dict((k, np.dtype(v)) for k, v in typemap.iteritems())

def to_dtype(type):
    if type.is_struct:
        fields = [(field_name, to_dtype(field_type))
                      for field_name, field_type in type.fields]
        return np.dtype(fields, align=not type.packed)
    elif type.is_array and type.ndim == 1:
        return to_dtype(type.dtype)
    elif type in typemap:
        return typemap[type]
    elif type.is_int:
        name = 'int' if type.signed else 'uint'
        return np.dtype(getattr(np, name + str(type.itemsize * 8)))
    else:
        raise ValueError("Cannot convert '%s' to numpy type" % (type,))


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import sys, os.path

_importlist = "VisitorBase parse check".split()

root = os.path.join(os.path.dirname(__file__))
common_path = os.path.join(root, 'common')

#------------------------------------------------------------------------
# ASDL Processing
#------------------------------------------------------------------------

class ASDLProcessor(object):
    """
    Allow pre-processing of ASDL source (str) and post-processing of the
    resulting ASDL tree.
    """

    def preprocess(self, asdl_source):
        return asdl_source

    def postprocess(self, asdl_tree):
        return asdl_tree

#------------------------------------------------------------------------
# Parse ASDL Schemas
#------------------------------------------------------------------------

class ASDLParser(object):
    """
    ASDL parser that accepts string inputs. Defers to the given ASDL module
    implementation.
    """

    def __init__(self, asdlmod, asdl_processor):
        self.asdlmod = asdlmod
        self.asdl_processor = asdl_processor

    def parse(self, buf):
        """
        Parse an ASDL string.
        """
        buf = self.asdl_processor.preprocess(buf)

        scanner = self.asdlmod.ASDLScanner()
        parser = self.asdlmod.ASDLParser()

        tokens = scanner.tokenize(buf)
        try:
            asdl_tree = parser.parse(tokens)
        except self.asdlmod.ASDLSyntaxError as err:
            raise ValueError("Error while parsing schema: %s" % (err,))
            # print(err)
            # lines = buf.split("\n")
            # print((lines[err.lineno - 1])) # lines starts at 0, files at 1

        asdl_tree = self.asdl_processor.postprocess(asdl_tree)
        return asdl_tree


    def check(self, mod, schema_name):
        """
        Check the validity of an ASDL parse.
        """
        v = self.asdlmod.Check()
        v.visit(mod)

        for t in v.types:
            if t not in mod.types and not t in self.asdlmod.builtin_types:
                v.errors += 1
                uses = ", ".join(v.types[t])
                print(("Undefined type %s, used in %s" % (t, uses)))

        if v.errors:
            raise ValueError(
                "Errors found while checking ASDL schema %r: %s" % (
                                             schema_name, v.errors))


#------------------------------------------------------------------------
# Load ASDL Schemas
#------------------------------------------------------------------------

class ASDLLoader(object):
    """
    Load a schema given an ASDLParser and an ASDL schema as a string.
    """

    def __init__(self, parser, schema_str, schema_name):
        self.parser = parser
        self.schema_str = schema_str
        self.schema_name = schema_name

    def load(self):
        asdl = self.parser.parse(self.schema_str)
        self.parser.check(asdl, self.schema_name)
        return asdl


def load(schema_name, schema_str, asdlmod, asdl_processor=None):
    asdl_processor = asdl_processor or ASDLProcessor()
    parser = ASDLParser(asdlmod, asdl_processor)
    loader = ASDLLoader(parser, schema_str, schema_name)
    return parser, loader

#------------------------------------------------------------------------
# Get ASDL implementation
#------------------------------------------------------------------------

def get_asdl_pydir():
    major, minor = sys.version_info[0], sys.version_info[1]

    # Assumes that specific-path and common-path are a subdirectory
    # Build an absolute module path.
    prefix = __name__.rsplit('.', 1)

    # The else-case is for running tests in the current directory
    base = (prefix[0] + '.') if len(prefix) > 1 else ''
    dir = 'py%d_%d' % (major, minor)

    return base, dir


def _get_asdl_depending_on_version():
    """
    Return Python ASDL implementation depending on the Python version.
    """
    use_abs_import = 0

    base, dir = get_asdl_pydir()
    modname = base + dir + '.asdl'
    try:
        # try to import from version specific directory
        mod = __import__(modname, fromlist=_importlist, level=use_abs_import)
    except ImportError:
        # fallback to import from common directory
        dir = 'common'
        modname = base + dir + '.asdl'
        mod = __import__(modname, fromlist=_importlist)

    return mod


def load_pyschema(filename):
    """
    Load ASDL from the version-specific directory if the schema exists in
    there, otherwise from the 'common' subpackage.

    Returns a two-tuple (ASDLParser, ASDLLoader).
    """
    base, dir = get_asdl_pydir()
    version_specific_path = os.path.join(root, dir)

    srcfile = os.path.join(version_specific_path, filename)
    if not os.path.exists(srcfile):
        srcfile = os.path.join(common_path, filename)
        from numba.asdl.common import asdl as asdlmod
    else:
        asdlmod = _get_asdl_depending_on_version()

    asdl_str = open(srcfile).read()
    return load(filename, asdl_str, asdlmod)

#------------------------------------------------------------------------
# Globals
#------------------------------------------------------------------------

python_parser, python_loader = load_pyschema("Python.asdl")

# Python version-specific parsed ASDL
python_asdl = python_loader.load()

# Python version-specific asdl implementation
pyasdl = python_parser.asdlmod



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import sys

from .testutils import generate_in_memory
from .. import generator, astgen, naming

def load_testschema1():
    file_allocator = generate_in_memory("testschema1.asdl", astgen.codegens)
    m = generator.generate_module(file_allocator, naming.nodes + '.py')
    return m

def test_ast_generation():
    """
    >>> test_ast_generation()
    """
    m = load_testschema1()

    # Nonterminals
    assert issubclass(m.root, m.AST)
    assert issubclass(m.expr, m.AST)

    # Products
    assert issubclass(m.myproduct, m.AST)

    # Terminals
    assert issubclass(m.Ham, m.root)
    assert issubclass(m.Foo, m.root)

    assert issubclass(m.SomeExpr, m.expr)
    assert issubclass(m.Bar, m.expr)

def test_ast_attributes():
    """
    >>> test_ast_attributes()
    """
    m = load_testschema1()

    assert m.root._attributes == ('foo', 'bar')

def test_valid_node_instantiation():
    """
    >>> test_valid_node_instantiation()
    Foo(leaf=Bar(e1=SomeExpr(n=10), e2=SomeExpr(n=11)))
    Bar(e1=SomeExpr(n=10), e2=None)
    Product(p=myproduct(foo=SomeExpr(n=10), bar=12))
    """
    m = load_testschema1()

    # Valid
    e1 = m.SomeExpr(10)
    e2 = m.SomeExpr(11)

    # Sum
    print(m.Foo(m.Bar(e1, e2)))
    print(m.Bar(e1, None))

    # Product
    print(m.Product(m.myproduct(e1, 12)))

def test_invalid_node_instantiation():
    """
    >>> m = load_testschema1()
    >>> e2 = m.SomeExpr(10)
    >>> m.Foo()
    Traceback (most recent call last):
      ...
    TypeError: ...
    >>> m.Bar(None, e2)
    Traceback (most recent call last):
      ...
    ValueError: Invalid type for attribute 'Bar.e1', expected instance of type(s) ('expr',) (got 'NoneType').
    >>> m.Product(e2)
    Traceback (most recent call last):
        ...
    ValueError: Invalid type for attribute 'Product.p', expected instance of type(s) ('myproduct',) (got 'SomeExpr').
    """


if __name__ == '__main__':
    import doctest
    optionflags = doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
    try:
        import cython
    except ImportError:
        print("Skipping test, cython not installed")
    else:
        sys.exit(0 if doctest.testmod(optionflags=optionflags).failed == 0 else 1)


# -*- coding: utf-8 -*-

"""
Inferface for our typesystems.

Some requirements:

  * The typesystem must allow us to switch between low-level representations.
    For instance, we may want to represent an array as a NumPy ndarray, a
    Py_buffer, or some other representation.

  * The sizes of atom types (e.g. int) must be easily customizable. This allows
    an interpreter to switch sizes to simulate different platforms.

  * Type representations and sizes, must be overridable without
    reimplementing the type. E.g. an atom type can be sliced to create
    an array type, which should be separate from its low-level representation.

  * Types should map easily between type domains of the same level, e.g.
    between the low-level numba and llvm types.
    Ideally this forms an isomorphism, which is precluded by ctypes and
    numpy type systems::

       >>> ctypes.c_longlong
       <class 'ctypes.c_long'>

  * No type system but our own should be entrenched in any part of the
    codebase, including the code generator.

  * Sets of type constructors (type universes) must be composable.
    For instance, we may want to extend a low-level type systems of ints
    and pointers with objects to yield a type universe supporting
    both constructs.

  * Universes must be re-usable across type-systems. Types of universes
    represent abstract type concepts, and the type-systems give meaning
    and representation to values of those types.

  * Types must be immutable and consed, i.e.
    ts.pointer(base) is ts.pointer(base) must always be True

  * The use of a certain typesystem must suggest at which level the
    corresponding terms operate.

  * Conversion code should be written with minimal effort:

        - unit types should map immediately between domains of the same level
        - parametric types should naturally re-construct in domains of the same level

  * Converting a type to a lower-level domain constitutes a one-way
    conversion. This should, where possible, constitute a lowering in the
    same domain followed by a conversion. E.g.:

        def numba_complex_to_llvm(type):
            return to_llvm(lower_complex(type))

  * Type constructors must be substitutable. E.g. an external user may
    want to construct a universe where type construction is logged, or
    where special type checking is performed, disallowing certain compositions.
"""

from __future__ import print_function, division, absolute_import

import ctypes
import struct as struct_
import weakref
import keyword
from functools import partial

import numba
from numba.utils import is_builtin

reserved = set(['bool', 'int', 'long', 'float', 'complex',
                'string', 'struct', 'array']).__contains__

def tyname(name):
    return name + "_" if reserved(name) else name

__all__ = [
    "TypeSystem", "Type", "ConstantTyper", "Conser", "TypeConser",
    "get_conser", "consing",
]

native_pointer_size = struct_.calcsize('@P')

if struct_.pack('i', 1)[0] == '\1':
    nbo = '<' # little endian
else:
    nbo = '>' # big endian

if numba.PY3:
    map = lambda f, xs, map=map: list(map(f, xs))

class TypeSystem(object):

    def __init__(self, universe, promote=None, typeof=None, converters=None):
        self.universe = universe

        # Find the least general type that subsumes both given types
        # t1 -> t2 -> t3
        self.promote = promote

        # Assign types to Python constants (arbitrary python values)
        self.typeof = typeof
        self.from_python = typeof # TODO: Remove

        # Convert between type domains
        self.converters = converters or {}

    def convert(self, codomain_name, type):
        convert = self.converters[codomain_name]
        return convert(type)

    def __getattr__(self, attr):
        return getattr(self.universe, attr)

    def __repr__(self):
        return "TypeSystem(%s, %s, %s, %s)" % (self.universe.domain_name,
                                               self.promote, self.typeof,
                                               self.converters)

#------------------------------------------------------------------------
# Typing of Constants
#------------------------------------------------------------------------

class ConstantTyper(object):

    def __init__(self, universe, typetable, handler_table):
        "Initialize to the given type universe"
        self.universe = universe
        self.typetable = typetable          # type(constant) -> type
        self.handler_table = handler_table  # type(constant) -> handler

    def typeof(self, value):
        """
        Get a concrete type given a python value.
        Return None f this ConstantTyper cannot type the constant
        """
        if type(value) in self.typetable:
            return self.typetable[type(value)]
        elif type(value) in self.handler_table:
            return self.handler_table[type(value)](self.universe, value)
        else:
            for cls in self.handler_table:
                if isinstance(value, cls):
                    return self.handler_table[cls](self.universe, value)

            return None

#------------------------------------------------------------------------
# Type Conversion between type domains
#------------------------------------------------------------------------

def get_ctor(codomain, kind):
    name = tyname(kind)
    if not hasattr(codomain, name):
        raise AttributeError(
            "Codomain '%s' has no attribute '%s'" % (codomain, name))
    return getattr(codomain, name)

def convert_unit(domain, codomain, type):
    return get_ctor(codomain, type.typename)

def convert_para(domain, codomain, type, coparams):
    return get_ctor(codomain, type.kind)(*coparams) # Construct type in codomain

# ______________________________________________________________________

class TypeConverter(object):
    """
    Map types between type universes.
    """

    def __init__(self, domain, codomain,
                 convert_unit=convert_unit, convert_para=convert_para):
        self.domain, self.codomain = domain, codomain
        self.convert_unit = partial(convert_unit, domain, codomain)
        self.convert_para = partial(convert_para, domain, codomain)
        self.partypes = weakref.WeakKeyDictionary()

    def convert(self, type):
        "Return an LLVM type for the given type."
        if isinstance(type, (tuple, list)):
            return tuple(map(self.convert, type))
        elif not isinstance(type, Type):
            return type
        elif type.is_unit:
            return self.convert_unit(type)
        else:
            return self.convert_parametrized(type)

    def convert_parametrized(self, type):
        # if type in self.partypes: # TODO: Check for type mutability
        #     return self.partypes[type]

        # Construct parametrized type in codomain
        result = self.convert_para(type, map(self.convert, type.params))

        # self.partypes[type] = result
        return result

    def __repr__(self):
        return "TypeConverter(%s -> %s)" % (self.domain.domain_name,
                                            self.codomain.domain_name)

#------------------------------------------------------------------------
# Type Classes
#------------------------------------------------------------------------

def add_flags(obj, flags):
    for flag in flags:
        setattr(obj, "is_" + flag, True)

class Type(object):
    """
    Base of all types.
    """
    metadata = None

    def __init__(self, kind, *params, **kwds):
        self.kind = kind    # Type kind

        # don't call this 'args' since we already use that in function
        self.params = list(params)
        self.is_unit = kwds.get("is_unit", False)
        if self.is_unit:
            self.typename = params[0]
        else:
            self.typename = kind

        # Immutable metadata
        self.metadata = kwds.get("metadata", frozenset())
        self._metadata = self.metadata and dict(self.metadata)

    @classmethod
    def unit(cls, kind, name, flags=(), **kwds):
        """
        Nullary type constructor creating the most elementary of types.
        Does not compose any other type.
        """
        type = cls(kind, name, is_unit=True,
                   metadata=frozenset(kwds.items()))
        add_flags(type, flags)
        type.flags = flags
        return type

    @classmethod
    def default_args(cls, args, kwargs):
        "Add defaults to a given args tuple for type construction"
        return args

    def __repr__(self):
        if self.is_unit:
            return self.params[0].rstrip("_")
        else:
            return "%s(%s)" % (self.kind, ", ".join(map(str, self.params)))

    def __getattr__(self, attr):
        if attr.startswith("is_"):
            return self.kind == attr[3:]
        elif self.metadata and attr in self._metadata:
            return self._metadata[attr]
        raise AttributeError( attr)

#------------------------------------------------------------------------
# Type Memoization
#------------------------------------------------------------------------

class Conser(object):
    """
    Conser: constructs new objects only when not already available.
    Objects are weakreffed to sanitize memory consumption.

    This allows the objects to be compared by and hashed on identity.
    """

    __slots__ = ("constructor", "_entries")

    def __init__(self, constructor):
        self._entries = weakref.WeakValueDictionary()
        self.constructor = constructor

    def get(self, *args):
        args = tuple(tuple(arg) if isinstance(arg, list) else arg
                         for arg in args)
        result = self._entries.get(args)
        if result is None:
            result = self.constructor(*args)
            self._entries[args] = result

        return result

class TypeConser(object):

    def __init__(self, type):
        assert isinstance(type, type), type
        assert issubclass(type, Type), type.__mro__
        self.type = type
        self.conser = Conser(type)

    def get(self, *args, **kwargs):
        # Add defaults to the arguments to ensure consing correctness
        args = self.type.default_args(args, kwargs)
        return self.conser.get(*args)

def get_conser(ctor):
    if isinstance(ctor, type) and issubclass(ctor, Type):
        return TypeConser(ctor) # Use a conser that tracks defaults
    else:
        return Conser(ctor)

def consing(ctor):
    return get_conser(ctor).get



import sys
import numba
from numba.testing.test_support import autojit_py3doc
# NOTE: See also issues/test_issue_56

autojit_py3doc = autojit_py3doc(warn=False, warnstyle='simple')

@autojit_py3doc
def test_bitwise_and(a, b):
    """
    >>> test_bitwise_and(0b01, 0b10)
    0
    >>> test_bitwise_and(0b01, 0b11)
    1

    >>> test_bitwise_and(0b01, 2.0)
    Traceback (most recent call last):
        ...
    NumbaError: 27:15: Expected an int, or object, or bool

    >>> test_bitwise_and(2.0, 0b01)
    Traceback (most recent call last):
        ...
    NumbaError: 27:11: Expected an int, or object, or bool

    """
    return a & b

@autojit_py3doc
def test_bitwise_or(a, b):
    """
    >>> test_bitwise_or(0b00, 0b00)
    0
    >>> test_bitwise_or(0b00, 0b01)
    1
    >>> test_bitwise_or(0b10, 0b00)
    2
    >>> test_bitwise_or(0b01, 0b10)
    3
    >>> test_bitwise_or(0b01, 0b11)
    3

    >>> test_bitwise_or(0b01, 2.0)
    Traceback (most recent call last):
        ...
    NumbaError: 54:15: Expected an int, or object, or bool

    >>> test_bitwise_or(2.0, 0b01)
    Traceback (most recent call last):
        ...
    NumbaError: 54:11: Expected an int, or object, or bool

    """
    return a | b


@autojit_py3doc
def test_bitwise_xor(a, b):
    """
    >>> test_bitwise_xor(0b00, 0b00)
    0
    >>> test_bitwise_xor(0b00, 0b01)
    1
    >>> test_bitwise_xor(0b10, 0b00)
    2
    >>> test_bitwise_xor(0b01, 0b10)
    3
    >>> test_bitwise_xor(0b01, 0b11)
    2

    >>> test_bitwise_xor(0b01, 2.0)
    Traceback (most recent call last):
        ...
    NumbaError: 82:15: Expected an int, or object, or bool

    >>> test_bitwise_xor(2.0, 0b01)
    Traceback (most recent call last):
        ...
    NumbaError: 82:11: Expected an int, or object, or bool

    """
    return a ^ b

@autojit_py3doc
def test_shift_left(a, b):
    """
    >>> test_shift_left(5, 2)
    20
    >>> test_shift_left(-5, 2)
    -20
    """
    return a << b

@autojit_py3doc
def test_shift_right(a, b):
    """
    >>> test_shift_right(20, 2)
    5
    >>> test_shift_right(-20, 2)
    -5
    """
    return a >> b

@autojit_py3doc
def test_invert(a):
    """
    >>> test_invert(5)
    -6
    >>> test_invert(-5)
    4
    """
    return ~a


numba.testing.testmod()



from numba.typesystem import *

assert int_.is_int
assert int_.is_numeric
assert long_.is_int
assert long_.is_numeric
assert not long_.is_long

assert float_.is_float
assert float_.is_numeric
assert double.is_float
assert double.is_numeric
assert not double.is_double

assert object_.is_object
assert list_(int_, 2).is_list
assert list_(int_, 2).is_object

assert function(void, [double]).is_function



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
from numba import autojit

@autojit(nopython=True)
def test_ord(s):
    return ord(s[1])

assert test_ord('hello') == 101


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import
import traceback

__all__ = ["NumbaError", "InternalError", "InvalidTemplateError"]

def format_pos(node):
    if node is not None and hasattr(node, 'lineno'):
        return format_postup((node.lineno, node.col_offset))
    else:
        return ""

def format_postup(tup):
    lineno, col_offset = tup
    return "%s:%s: " % (lineno, col_offset)

class NumbaError(Exception):
    "Some error happened during compilation"

    def __init__(self, node, msg=None, *args, **kwds):
        if msg is None:
            node, msg = None, node

        self.node = node
        self.msg = msg
        self.args = args
        self.has_report = kwds.get("has_report", False)

    def __str__(self):
        try:
            if self.has_report:
                return self.msg.strip()
            pos = format_pos(self.node)
            msg = "%s%s %s" % (pos, self.msg, " ".join(map(str, self.args)))
            return msg.rstrip()
        except:
            traceback.print_exc()
            return "<internal error creating numba error message>"


class InternalError(Exception):
    "Indicates a compiler bug"

class _UnknownAttribute(Exception):
    pass

class InvalidTemplateError(Exception):
    "Raised for invalid template type specifications"

class UnpromotableTypeError(TypeError):
    "Raised when we can't promote two given types"
    def __str__(self):
        return "Cannot promote types %s and %s" % self.args[0]



# -*- coding: utf-8 -*-

"""
Generate Python classes and Cython pxd files.
"""

from __future__ import print_function, division, absolute_import

from . import generator, formatting
from numba.asdl.schema import verify_schema_keywords

class Class(object):
    def __init__(self, name, base, doc, fields=(), attributes=None):
        self.name = name
        self.base = base
        self.doc = doc
        self.fields = fields
        self.attributes = attributes

def doc_sumtype(sumtype, fields):
    if fields:
        return "%s(%s)" % (sumtype, formatting.format_fields(fields))
    return sumtype

class ClassCodegen(generator.SimpleCodegen):
    """
    Generate Python AST nodes.
    """

    def __init__(self, out_filename, preamble, Class, rootclass):
        super(ClassCodegen, self).__init__(out_filename)
        self.preamble = preamble
        self.rootclass = rootclass
        self.Class = Class

    def emit_preamble(self, emitter, schema):
        verify_schema_keywords(schema)
        emitter.emit(self.preamble)
        emitter.emit(self.rootclass)

    def emit_nonterminal(self, emitter, schema, rulename, rule):
        "Emit code for a rule (a nonterminal)"
        nonterminal_fields = schema.attributes[rulename]
        attrs = tuple(map(repr, (field.name for field in nonterminal_fields)))

        alignment = 5 + len(rulename)
        sep = "\n%s| " % (" " * alignment)
        doc = "%s = %s" % (rulename, sep.join(
            doc_sumtype(t, schema.types[t]) for t in rule.fields))

        c = self.Class(
            rulename,
            self.rootclass.name,
            doc=doc,
            # Attributes: Use None instead of empty tuple to inherit from base
            attributes=attrs or None,
        )
        emitter.emit(c)

    def emit_product(self, emitter, schema, rulename, rule):
        doc = "%s = (%s)" % (rulename, formatting.format_fields(rule.fields))
        emitter.emit(self.Class(rulename, self.rootclass.name,
                                doc=doc, fields=rule.fields))

    def emit_sum(self, emitter, schema, rulename, rule, sumtype):
        fields = schema.types[sumtype]
        doc = doc_sumtype(sumtype, fields)
        emitter.emit(self.Class(sumtype, rulename, doc=doc, fields=fields))



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

import numba
from numba import autojit
import numpy as np

def test(func, expect):
    N = 10
    A = np.empty(N, dtype=np.float32)
    A[...] = 2.5
    func(A)
    assert np.allclose(A, expect)

@autojit
def forward1(A):
    for i in numba.prange(10):
        A[i] = i

@autojit
def forward2(A):
    for i in numba.prange(1, 5):
        A[i] = i

@autojit
def forward3(A):
    for i in numba.prange(1, 8, 3):
        A[i] = i

@autojit
def backward1(A):
    for i in numba.prange(9, 2, -3):
        A[i] = i

@autojit
def backward2(A):
    for i in numba.prange(1, 5, -1):
        A[i] = i

@autojit
def empty_assign():
    i = 14
    for i in numba.prange(10, 4):
        pass
    return i

@autojit
def last_value():
    for i in numba.prange(10):
        pass
    print("after loop", i)
    return i

def run():
    test(forward1,
         [ 0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9.])
    test(forward2,
         [ 2.5,  1.,   2.,   3.,   4.,   2.5,  2.5,  2.5,  2.5,  2.5])
    test(forward3,
         [ 2.5,  1.,   2.5,  2.5,  4.,   2.5,  2.5,  7.,   2.5,  2.5])
    test(backward1,
         [ 2.5,  2.5,  2.5,  3.,   2.5,  2.5,  6.,   2.5,  2.5,  9. ])
    test(backward2,
         [ 2.5,  2.5,  2.5,  2.5,  2.5,  2.5,  2.5,  2.5,  2.5,  2.5])

    assert empty_assign() == 14

    last = last_value()
    print(last, 9)

if __name__ == '__main__':
    run()


# -*- coding: utf-8 -*-

"""
Generate IR utilities from ASDL schemas.
"""

from __future__ import print_function, division, absolute_import

import os
import types
import codecs
import textwrap

from numba import PY3
from numba.asdl import asdl
from numba.asdl import schema
from numba.asdl import processor
from numba.asdl.asdl import pyasdl

if PY3:
    import io
    BytesIO = io.BytesIO
else:
    import cStringIO
    BytesIO = cStringIO.StringIO

root = os.path.dirname(os.path.abspath(__file__))

#------------------------------------------------------------------------
# Entry Points
#------------------------------------------------------------------------

# TODO: This needs to be parametrized when this is externalized
asdl_import_path = [os.path.join(root, os.pardir)]
import_processor = processor.ImportProcessor(pyasdl, asdl_import_path)

def parse(schema_name, schema_def, asdl_processor=import_processor):
    """
    Parse a schema given a schema name and schema string (ASDL string).
    """
    asdl_tree = get_asdl(schema_name, schema_def, asdl_processor)
    schema_instance = schema.build_schema(asdl_tree)
    return asdl_tree, schema_instance

def generate(schema_name, schema_str, codegens, file_allocator):
    """
    schema: ASDL schema (str)
    """
    asdl_tree, schema_instance = parse(schema_name, schema_str)

    for codegen in codegens:
        emitter = codegen.make_code_emitter(file_allocator)
        codegen.generate(emitter, asdl_tree, schema_instance)

def generate_from_file(schema_filename, codegens, output_dir):
    """
    Generate code files for the given schema, code generators and output
    directory.

    Returns a file allocator with the open disk files.
    """
    schema_name = os.path.basename(schema_filename)
    schema_str = open(schema_filename).read()
    file_allocator = DiskFileAllocator(output_dir)
    generate(schema_name, schema_str, codegens, file_allocator)
    return file_allocator

def generate_module(file_allocator, name):
    """
    Generate an in-memory module from a generated Python implementation.
    """
    assert name in file_allocator.allocated_files

    f = file_allocator.allocated_files[name]
    f.seek(0)
    data = f.read()

    modname, _ = os.path.splitext(name)

    d = {}
    eval(compile(data, name, "exec"), d, d)
    m = types.ModuleType(modname)
    vars(m).update(d)

    return m

#------------------------------------------------------------------------
# Code Generator Utilities
#------------------------------------------------------------------------

def get_asdl(schema_name, schema, asdl_processor=None):
    parser, loader = asdl.load(schema_name, schema, pyasdl,
                               asdl_processor=asdl_processor)
    asdl_tree = loader.load()
    return asdl_tree

def is_simple(sum):
    """
    Return True if a sum is a simple.

    A sum is simple if its types have no fields, e.g.
    unaryop = Invert | Not | UAdd | USub
    """
    for t in sum.types:
        if t.fields:
            return False
    return True

#------------------------------------------------------------------------
# File Handling
#------------------------------------------------------------------------

StreamWriter = codecs.getwriter('UTF-8')

class FileAllocator(object):

    def __init__(self, output_dir=None):
        self.output_dir = output_dir

        # file_name -> file
        self.allocated_files = {}

    def open_sourcefile(self, name):
        "Allocate a file and save in in allocated_files"

    def close(self):
        for file in self.allocated_files.itervalues():
            file.close()

        self.allocated_files.clear()

class DiskFileAllocator(FileAllocator):

    def open_sourcefile(self, name):
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

        filename = os.path.join(self.output_dir, name)
        file = codecs.open(filename, 'w', encoding='UTF-8')
        self.allocated_files[name] = file
        return file

class MemoryFileAllocator(FileAllocator):

    def open_sourcefile(self, name):
        file = StreamWriter(BytesIO())
        self.allocated_files[name] = file
        return file

#------------------------------------------------------------------------
# Code Generator Interface
#------------------------------------------------------------------------

class CodeEmitter(object):

    def __init__(self, outfile):
        self.outfile = outfile

    def emit(self, s):
        self.outfile.write(unicode(s))


class Codegen(object):
    """
    Interface for code generators.
    """

    def __init__(self, out_filename):
        self.out_filename = out_filename

    def make_code_emitter(self, file_allocator):
        outfile = file_allocator.open_sourcefile(self.out_filename)
        return CodeEmitter(outfile)

    def generate(self, emitter, asdl_tree, schema_instance):
        """
        Generate code for the given asdl tree. The ASDL tree is accompanied
        by a corresponding schema.Schema, which is easier to deal with.
        """


class UtilityCodegen(Codegen):
    """
    Generate some utility code:

        UtilityCode("foo.py", "my python code")
    """

    def __init__(self, out_filename, utility_code):
        super(UtilityCodegen, self).__init__(out_filename)
        self.utility_code = utility_code

    def generate(self, emitter, asdl_tree, schema_instance):
        emitter.emit(self.utility_code)


class SimpleCodegen(Codegen):

    def generate(self, emitter, asdl_tree, schema):
        self.emit_preamble(emitter, schema)

        for rulename, rule in schema.dfns.iteritems():
            if rule.is_sum:
                self.emit_nonterminal(emitter, schema, rulename, rule)

        for rulename, rule in schema.dfns.iteritems():
            if rule.is_product:
                self.emit_product(emitter, schema, rulename, rule)

        for rulename, rule in schema.dfns.iteritems():
            if rule.is_sum:
                for sumtype in rule.fields:
                    self.emit_sum(emitter, schema, rulename, rule, sumtype)

    def emit_preamble(self, emitter, schema):
        pass

    def emit_nonterminal(self, emitter, schema, rulename, rule):
        pass

    def emit_product(self, emitter, schema, rulename, rule):
        pass

    def emit_sum(self, emitter, schema, rulename, rule, sumtype):
        pass

if __name__ == '__main__':
    schema_def = textwrap.dedent("""
        module MyModule version "0.1"
        {
            mod = Module(object leaf1, object leaf2)
                | Foo(object leaf)

            foo = Add | Mul
            expr = X | Y
                 attributes (int lineno)

            alias = (int foo, int bar)
        }
    """)
    asdl_tree = get_asdl("MyASDL.asdl", schema_def)
    print("asdl", asdl_tree)
    print ("-------------")
    s = schema.build_schema(asdl_tree)
    print("dfns", s.dfns)
    print("types", s.types)



from __future__ import print_function, division, absolute_import

import numpy as np
from numba import autojit

def test(a, b):
    r = np.empty(3, dtype=bool)
    for i in range(len(a)):
        r[i] = a[i] != b[i]
    return r
test_nb = autojit(test)

a = np.arange(3, dtype=complex)
b = np.arange(3, dtype=complex)
b[1] += 1j

assert np.array_equal(test(a, b), test_nb(a, b))



# -*- coding: utf-8 -*-
"""
Resolve calls to math functions.

During type inference this annotates math calls, and during
final specialization it produces LLVMIntrinsicNode and MathCallNode
nodes.
"""
from __future__ import print_function, division, absolute_import

import math
import cmath
import collections
try:
    import __builtin__ as builtins
except ImportError:
    import builtins

import numpy as np

from numba import *
from numba import nodes
from numba.symtab import Variable
from numba.typesystem import get_type, rank
from numba.type_inference.modules import utils

#----------------------------------------------------------------------------
# Utilities
#----------------------------------------------------------------------------

register_math_typefunc = utils.register_with_argchecking

def binop_type(typesystem, x, y):
    "Binary result type for math operations"
    x_type = get_type(x)
    y_type = get_type(y)
    return typesystem.promote(x_type, y_type)

#----------------------------------------------------------------------------
# Determine math functions
#----------------------------------------------------------------------------

# sin(double), sinf(float), sinl(long double)
mathsyms = [
    'sin',
    'cos',
    'tan',
    'sqrt',
    'acos',
    'asin',
    'atan',
    'sinh',
    'cosh',
    'tanh',
    'asinh',
    'acosh',
    'atanh',
    'log',
    'log2',
    'log10',
    'erfc',
    'floor',
    'ceil',
    'exp',
    'exp2',
    'expm1',
    'rint',
    'log1p',
]

n_ary_mathsyms = {
    'hypot'     : 2,
    'atan2'     : 2,
    'logaddexp' : 2,
    'logaddexp2': 2,
    'pow'       : (2, 3),
}

math2ufunc = {
    'asin' : 'arcsin',
    'acos' : 'arccos',
    'atan' : 'arctan',
    'asinh': 'arcsinh',
    'acosh': 'arccosh',
    'atanh': 'arctanh',
    'atan2': 'arctan2',
    'pow'  : 'power',
}

ufunc2math = dict((v, k) for k, v in math2ufunc.items())

#----------------------------------------------------------------------------
# Math Type Inferers
#----------------------------------------------------------------------------

# TODO: Move any rewriting parts to lowering phases

def mk_infer_math_call(default_result_type):
    def infer(typesystem, call_node, *args):
        "Resolve calls to llvmmath math calls"
        # signature is a generic signature, build a correct one
        type = reduce(typesystem.promote, map(get_type, call_node.args))

        if type.is_numeric and rank(type) < rank(default_result_type):
            type = default_result_type
        elif type.is_array and type.dtype.is_int:
            type = typesystem.array(double, type.ndim)

        call_node.args[:] = nodes.CoercionNode.coerce(call_node.args, type)

        # TODO: Remove the abuse below
        nodes.annotate(typesystem.env, call_node, is_math=True)
        call_node.variable = Variable(type)
        return call_node

    return infer

infer_math_call = mk_infer_math_call(double)
infer_cmath_call = mk_infer_math_call(complex128)

# ______________________________________________________________________
# abs()

def abs_(typesystem, node, x):
    from . import builtinmodule

    argtype = get_type(x)
    nodes.annotate(typesystem.env, node, is_math=True)

    if argtype.is_array and argtype.dtype.is_numeric:
        # Handle np.abs() on arrays
        dtype = builtinmodule.abstype(argtype.dtype)
        result_type = argtype.add('dtype', dtype)
        node.variable = Variable(result_type)
    else:
        node = builtinmodule.abs_(typesystem, node, x)

    return node

register_math_typefunc(1)(abs_, np.abs)

#----------------------------------------------------------------------------
# Register Type Functions
#----------------------------------------------------------------------------

def register_math(infer_math_call, nargs, value):
    register = register_math_typefunc(nargs)
    register(infer_math_call, value)

def npy_name(name):
    return math2ufunc.get(name, name)

id_name = lambda x: x

# ______________________________________________________________________

def reg(mod, register, getname):
    """Register all functions listed in mathsyms and n_ary_mathsyms"""
    nargs = lambda f: n_ary_mathsyms.get(f, 1)
    for symname in mathsyms + list(n_ary_mathsyms):
        if hasattr(mod, getname(symname)):
            register(nargs(symname), getattr(mod, getname(symname)))

reg(math, partial(register_math, infer_math_call), id_name)
reg(cmath, partial(register_math, infer_cmath_call), id_name)
reg(np, partial(register_math, infer_math_call), npy_name)


import llvm.core

from numba.typesystem.itypesystem import consing, tyname
from numba.typesystem import universe
# from llvmmath.ltypes import l_longdouble

domain_name = "llvm"

#------------------------------------------------------------------------
# Helpers
#------------------------------------------------------------------------

def get_target_triple():
    target_machine = llvm.ee.TargetMachine.new()
    is_ppc = target_machine.triple.startswith("ppc")
    is_x86 = target_machine.triple.startswith("x86")
    return is_ppc, is_x86

def lbool():
    return llvm.core.Type.int(1)

def lint(name, itemsize):
    if name == "bool":
        return lbool()

    return llvm.core.Type.int(itemsize * 8)

def lfloat(name, itemsize):
    if itemsize == 4:
        return llvm.core.Type.float()
    elif itemsize == 8:
        return llvm.core.Type.double()
    else:
        assert False, "long double is not supported"
        # return l_longdouble

size = universe.default_type_sizes.__getitem__

unittypes = {}
for typename in universe.int_typenames:
    unittypes[typename] = lint(typename, size(typename))
for typename in universe.float_typenames:
    unittypes[typename] = lfloat(typename, size(typename))
unittypes["void"] = llvm.core.Type.void()

globals().update((tyname(name), ty) for name, ty in unittypes.iteritems())

#------------------------------------------------------------------------
# Exposed types
#------------------------------------------------------------------------

# @consing # llvm types don't hash in python 3 in llvmpy 0.11.2
def struct_(fields, name=None, readonly=False, packed=False):
    if packed:
        struct = llvm.core.Type.packed_struct
    else:
        struct = llvm.core.Type.struct

    return struct([field_type for field_name, field_type in fields])

# @consing
def pointer(base_type):
    if base_type.kind == llvm.core.TYPE_VOID:
        base_type = llvm.core.Type.int(8)
    return llvm.core.Type.pointer(base_type)

def sized_pointer(base_type, size):
    return pointer(base_type)

# @consing
def function(rettype, argtypes, name=None, is_vararg=False):
    return llvm.core.Type.function(rettype, argtypes, is_vararg)

def array_(dtype, ndim, *args):
    from numba import environment, ndarray_helpers
    # TODO: this is gross, we need to pass in 'env'
    env = environment.NumbaEnvironment.get_environment()
    if env.crnt:
        return env.crnt.array.from_type(dtype)
    return ndarray_helpers.NumpyArray.from_type(dtype)

carray = llvm.core.Type.array


# -*- coding: utf-8 -*-

"""
Type defaults
"""

from __future__ import print_function, division, absolute_import

from . itypesystem import TypeConverter, TypeSystem
from . import promotion, constants, lowering
from . import numbatypes as numba_domain
from . import llvmtypes as llvm_domain
from . import ctypestypes as ctypes_domain

#------------------------------------------------------------------------
# Defaults initialization
#------------------------------------------------------------------------

def compose(f, g):
    return lambda x: f(g(x))

# ______________________________________________________________________
# Converters

def lowerer(table):
    return lowering.create_type_lowerer(table, numba_domain, numba_domain)

# Lowerers
lower = lowerer(lowering.default_numba_lowering_table).convert
ctypes_lower = lowerer(lowering.ctypes_lowering_table).convert

# Converters
to_llvm_converter = TypeConverter(numba_domain, llvm_domain)
to_ctypes_converter = TypeConverter(numba_domain, ctypes_domain)

# ...
to_llvm = to_llvm_converter.convert
to_ctypes = to_ctypes_converter.convert

converters = {
    "llvm": compose(to_llvm, lower),
    "ctypes": compose(to_ctypes, ctypes_lower),
}

# ______________________________________________________________________
# Typesystems

promote = promotion.get_default_promoter(numba_domain)
typeof = constants.get_default_typeof(numba_domain, promote)

numba_typesystem = TypeSystem(numba_domain, promote, typeof, converters)
llvm_typesystem = TypeSystem(llvm_domain, typeof=compose(to_llvm, typeof))
ctypes_typesystem = TypeSystem(ctypes_domain, typeof=compose(to_ctypes, typeof))



import numba
from numba import *
from numba.typesystem import typeset, promote
from numba.environment import NumbaEnvironment

def s(type):
    return type(type, type)

def typeset_matching():
    numeric = typeset.typeset([int_, longlong])
    n = numeric(numeric, numeric)
    f = numba.floating(numba.floating, numba.floating)

    signatures = [n, f, object_(object_, object_)]
    ts = typeset.typeset(signatures)

    assert ts.find_match(promote, [float_, float_]) == s(float_)
    assert ts.find_match(promote, [float_, double]) == s(double)
    # assert ts.find_match(promote, [longdouble, float_]) == s(longdouble)

    assert ts.find_match(promote, [int_, int_]) == s(int_)
    # assert ts.find_match(promote, [int_, longlong]) == s(longlong)
    # assert ts.find_match(promote, [short, int_]) == s(int_)

    # np.result_type(np.short, np.ulonglong) -> np.float64
    # assert ts.find_match(promote, [short, ulonglong]) is None

if __name__ == '__main__':
    typeset_matching()



# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from .types import *
from .itypesystem import *
from .closuretypes import *
from .templatetypes import *
from numba.exttypes.types.extensiontype import *
from numba.exttypes.types.methods import *

from . import numbatypes
from .numbatypes import *

from .ssatypes import *
# from typeset import *
from .typematch import *

from .universe import *
from .defaults import *
from .typeutils import *

__all__ = numbatypes.__all__



class Formatter(object):

    def format_stats(self, pattern, indent, stats):
        pattern = pattern + " " * indent
        return pattern.join(stats)

    def get_fields(self, fields, obj="self"):
        fieldnames = (str(field.name) for field in fields)
        return ["%s.%s" % (obj, name) for name in fieldnames]

class PythonFormatter(Formatter):

    def format_type(self, asdl_type):
        """
        ASDL's five builtin types are identifier, int, string, object, bool
        """
        type = str(asdl_type)

        defaults = {
            'identifier': 'str',
            'string': 'str',
        }

        return defaults.get(type, type)

class CythonFormatter(Formatter):

    def format_type(self, asdl_type):
        """
        ASDL's five builtin types are identifier, int, string, object, bool
        """
        type = str(asdl_type)

        defaults = {
            'identifier': 'str',
            'string': 'str',
            'bool': 'bint',
        }

        return defaults.get(type, type)

def format_fields(fields):
    return ", ".join("%s %s" % (f.type, f.name) for f in fields)

py_formatter = PythonFormatter()
cy_formatter = CythonFormatter()


# -*- coding: utf-8 -*-
from __future__ import print_function, division, absolute_import

from numba.decorators import jit
from numba import *
import numpy as np
import numpy.core.umath_tests as ut
from numba.vectorize import GUVectorize

def matmulcore(A, B, C):
    m, n = A.shape
    n, p = B.shape
    for i in range(m):
        for j in range(p):
            C[i, j] = 0
            for k in range(n):
                C[i, j] += A[i, k] * B[k, j]


def test_numba():
    jit_matmulcore = jit(argtypes=[f4[:,:], f4[:,:], f4[:,:]])(matmulcore)

    A = np.arange(16, dtype=np.float32).reshape(4, 4)
    B = np.arange(16, dtype=np.float32).reshape(4, 4)
    C = np.zeros(16, dtype=np.float32).reshape(4, 4)
    Gold = np.matrix(A) * np.matrix(B)

    jit_matmulcore(A, B, C)

    if (C != Gold).any():
        raise ValueError

def _test_gufunc(backend, target):
    gufunc = GUVectorize(matmulcore, '(m,n),(n,p)->(m,p)')
    gufunc.add(argtypes=[f4[:,:], f4[:,:], f4[:,:]])
    gufunc = gufunc.build_ufunc()

    matrix_ct = 1001 # an odd number to test thread/block division in CUDA
    A = np.arange(matrix_ct * 2 * 4, dtype=np.float32).reshape(matrix_ct, 2, 4)
    B = np.arange(matrix_ct * 4 * 5, dtype=np.float32).reshape(matrix_ct, 4, 5)

    C = gufunc(A, B)
    Gold = ut.matrix_multiply(A, B)

#     print(A)
#     print(B)
#    print(C)
#    print(Gold)
    assert np.allclose(C, Gold)

#
### test gufuncs
#
def array_expr_gufunc(A, B, C):
    m, n = A.shape
    n, p = B.shape
    for i in range(m):
        for j in range(p):
            result = (A[i, :] * B[:, j]).sum()
            # print result
            C[i, j] = result

def test_gufunc_array_expressions():
    gufunc = GUVectorize(array_expr_gufunc, '(m,n),(n,p)->(m,p)')
    gufunc.add(argtypes=[float_[:,:], float_[:,:], float_[:,:]])
    gufunc = gufunc.build_ufunc()

    matrix_ct = 10
    A = np.arange(matrix_ct * 2 * 4, dtype=np.float32).reshape(matrix_ct, 2, 4)
    B = np.arange(matrix_ct * 4 * 5, dtype=np.float32).reshape(matrix_ct, 4, 5)

    C = gufunc(A, B)
    Gold = ut.matrix_multiply(A, B)

    if (C != Gold).any():
        print(C)
        print(Gold)
        raise ValueError

def test_gufunc():
    #_test_gufunc('bytecode', 'cpu')
    _test_gufunc('ast', 'cpu')

def main():
    for i in range(10):
        test_numba()
        test_gufunc()
        test_gufunc_array_expressions()
    print('ok')

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Shorthands for type constructing, promotions, etc.
"""
from __future__ import print_function, division, absolute_import
import __future__

import inspect

from numba.typesystem import types, universe
from numba.typesystem.types import *

__all__ = [] # set below

integral = []
unsigned_integral = []
floating = []
complextypes = []
numeric = []
native_integral = []

domain_name = "numba"

ranking = ["bool", "int", "float", "complex", "object"]

def rank(type):
    return ranking.index(type.kind)

#------------------------------------------------------------------------
# All unit types
#------------------------------------------------------------------------

def unit(*args, **kwargs):
    ty = types.unit(*args, **kwargs)

    if ty.is_int:
        ty.signed = ty.typename in universe.signed
    if ty.is_int or ty.is_float:
        ty.itemsize = universe.default_type_sizes[ty.typename]

    # Add types to categories numeric, integral, floating, etc...
    if ty.is_int:
        integral.append(ty)
        if not ty.signed:
            unsigned_integral.append(ty)
        if universe.is_native_int(ty.typename):
            native_integral.append(ty)
    elif ty.is_float:
        floating.append(ty)

    if ty.is_numeric:
        numeric.append(ty)

    return ty

# Numeric types
char         = unit("int", "char",         flags=["numeric"])
uchar        = unit("int", "uchar",        flags=["numeric"])
short        = unit("int", "short",        flags=["numeric"])
ushort       = unit("int", "ushort",       flags=["numeric"])
int_         = unit("int", "int",          flags=["numeric"])
uint         = unit("int", "uint",         flags=["numeric"])
long_        = unit("int", "long",         flags=["numeric"])
ulong        = unit("int", "ulong",        flags=["numeric"])
longlong     = unit("int", "longlong",     flags=["numeric"])
ulonglong    = unit("int", "ulonglong",    flags=["numeric"])
int8         = unit("int", "int8",         flags=["numeric"])
int16        = unit("int", "int16",        flags=["numeric"])
int32        = unit("int", "int32",        flags=["numeric"])
int64        = unit("int", "int64",        flags=["numeric"])
uint8        = unit("int", "uint8",        flags=["numeric"])
uint16       = unit("int", "uint16",       flags=["numeric"])
uint32       = unit("int", "uint32",       flags=["numeric"])
uint64       = unit("int", "uint64",       flags=["numeric"])
size_t       = unit("int", "size_t",       flags=["numeric"])
npy_intp     = unit("int", "npy_intp",     flags=["numeric"])
Py_ssize_t   = unit("int", "Py_ssize_t",   flags=["numeric"])
Py_uintptr_t = unit("int", "Py_uintptr_t", flags=["numeric"])

float32      = unit("float", "float32",    flags=["numeric"])
float64      = unit("float", "float64",    flags=["numeric"])
# float128     = unit("float", "float128",   flags=["numeric"])
# float_, double, longdouble = float32, float64, float128
float_, double = float32, float64

complex64    = complex_(float32)
complex128   = complex_(float64)
# complex256   = complex_(float128)

bool_        = unit("bool", "bool", flags=["int", "numeric"])
null         = unit("null", "null", flags=["pointer"])
void         = unit("void", "void")

obj_type = lambda name: unit(name, name, flags=["object"])

# Add some unit types... (objects)
object_      = obj_type("object")
unicode_     = obj_type("unicode")
none         = obj_type("none")
ellipsis     = obj_type("ellipsis")
slice_       = obj_type("slice")
newaxis      = obj_type("newaxis")
range_       = obj_type("range")
string_      = unit("string", "string", flags=[#"object",
                                               "c_string"])
c_string_type = string_

complextypes.extend([complex64, complex128]) #, complex256])

tuple_of_obj       = tuple_(object_, -1)
list_of_obj        = list_(object_, -1)
dict_of_obj        = dict_(object_, object_, -1)

# ______________________________________________________________________

O = object_
b1 = bool_
i1 = int8
i2 = int16
i4 = int32
i8 = int64
u1 = uint8
u2 = uint16
u4 = uint32
u8 = uint64

f4 = float32
f8 = float64
# f16 = float128

c8 = complex64
c16 = complex128
# c32 = complex256

for name, value in list(globals().iteritems()): # TODO: Do this better
    if (not isinstance(value, __future__ ._Feature) and not
        inspect.ismodule(value) and not name.startswith("_")):
        __all__.append(name)